"""
Array (Tensor) functions
"""

from __future__ import annotations
from typing import Optional, Union, TypeVar, Sequence
import numpy
from returnn.tensor import Tensor, Dim
from ._backend import global_backend, get_backend_by_raw_tensor_type
from .types import RawTensorTypes

T = TypeVar("T")

__all__ = ["convert_to_tensor", "constant", "gather"]


def convert_to_tensor(
    value: Union[Tensor, T, RawTensorTypes],
    *,
    dims: Sequence[Dim] = None,
    dtype: Optional[str] = None,
    sparse_dim: Optional[Dim] = None,
    shape: Sequence[Dim] = None,
) -> Tensor[T]:
    """
    :param value: tensor, or scalar raw tensor or some other scalar value
    :param dims:
    :param dtype:
    :param sparse_dim:
    :param shape: alias for dims, for some older code
    :return: tensor
    """
    if isinstance(value, Tensor):  # fast path
        return value
    if isinstance(value, (int, float, complex, bool, str, numpy.number, numpy.ndarray)):
        backend = global_backend
    else:
        backend = get_backend_by_raw_tensor_type(value)
    if dims is None and shape is not None:
        dims = shape
    if dims is None:
        dims = ()
    return backend.convert_to_tensor(value=value, dims=dims, dtype=dtype, sparse_dim=sparse_dim)


constant = convert_to_tensor  # alias for some older code


# noinspection PyUnusedLocal
def gather(
    source: Tensor,
    *,
    indices: Union[Tensor, int],
    axis: Dim,
    clip_to_valid: bool = False,
) -> Tensor:
    """
    Gathers slices on a specified axis from the source using indices.
    If the source is of the shape ``[B,D,F1]``, and indices of shape ``[B,F2]``,
    this will yield output of the shape ``[B,F2,F1]`` where

    ``output[b,f2,f1] = source[b,indices[b,f2],f1]``

    (if ``D`` is the axis to gather from).
    In general, all shared axes of the input and the positions will be considered as batch-axes.

    The ``indices`` argument can also be an ``int``.
    In this case, this simply gives ``source[indices]`` one the specified ``axis``.

    :param source:
    :param indices: indices used to select the slices of the source from.
        If another layer, must be of type ``int32`` or ``int64``.
        Can also specify a constant ``int``.
    :param axis: The axis into which we gather the indices into
    :param clip_to_valid: if True, the indices will be clipped to the valid range of the input
        Also taking seq lengths into account.
    :return: layer
    """
    raise NotImplementedError



"""
Some datasets for artificially generated data.
"""

from __future__ import annotations

from typing import Optional, Sequence
import numpy
import sys
import typing

from returnn.util.basic import class_idx_seq_to_1_of_k, CollectionReadCheckCovered
from returnn.log import log

from .util.feature_extraction import ExtractAudioFeatures
from .util.vocabulary import *
from .audio import OggZipDataset  # noqa # for API compatibility
from .basic import Dataset, DatasetSeq, convert_data_dims
from .cached2 import CachedDataset2


class GeneratingDataset(Dataset):
    """
    Some base class for datasets with artificially generated data.
    """

    _input_classes = None
    _output_classes = None

    def __init__(self, input_dim, output_dim, num_seqs=float("inf"), **kwargs):
        """
        :param int|None input_dim:
        :param int|dict[str,int|(int,int)|dict] output_dim: if dict, can specify all data-keys
        :param int|float num_seqs:
        """
        super(GeneratingDataset, self).__init__(**kwargs)
        assert self.shuffle_frames_of_nseqs == 0

        self.num_inputs = input_dim
        output_dim = convert_data_dims(output_dim, leave_dict_as_is=False)
        if "data" not in output_dim and input_dim is not None:
            output_dim["data"] = (input_dim * self.window, 2)  # not sparse
        self.num_outputs = output_dim
        self.expected_load_seq_start = 0
        self._seq_order = None  # type: Optional[Sequence[int]]
        self._num_seqs = num_seqs
        self._total_num_seqs = num_seqs
        self.random = numpy.random.RandomState(1)
        self.reached_final_seq = False
        self.added_data = []  # type: typing.List[DatasetSeq]
        if self.seq_ordering in ("sorted", "sorted_reverse"):
            # For the dev/eval dataset, RETURNN automatically tries to sort them.
            # As this is not supported, just ignore it and reset it to the default order.
            self.seq_ordering = "default"

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :type epoch: int|None
        :param list[str]|None seq_list: predefined order via tags, doesn't make sense here
        :param list[int]|None seq_order: predefined order via indices, doesn't make sense here
        This is called when we start a new epoch, or at initialization.
        """
        super(GeneratingDataset, self).init_seq_order(epoch=epoch)
        assert seq_list is None and seq_order is None, (
            "predefined order doesn't make sense for %s" % self.__class__.__name__
        )
        self.random.seed(self._get_random_seed_for_epoch(epoch=epoch))
        if self._total_num_seqs == float("inf"):
            assert self.seq_ordering == "default"
            self._seq_order = None
            self._num_seqs = self._total_num_seqs
        else:
            self._seq_order = self.get_seq_order_for_epoch(epoch=epoch, num_seqs=self._total_num_seqs, get_seq_len=None)
            self._num_seqs = len(self._seq_order)
        self._num_timesteps = 0
        self.reached_final_seq = False
        self.expected_load_seq_start = 0
        self.added_data = []
        return True

    def _cleanup_old_seqs(self, seq_idx_end):
        i = 0
        while i < len(self.added_data):
            if self.added_data[i].seq_idx >= seq_idx_end:
                break
            i += 1
        del self.added_data[:i]

    def _check_loaded_seq_idx(self, seq_idx):
        if not self.added_data:
            raise Exception("no data loaded yet")
        start_loaded_seq_idx = self.added_data[0].seq_idx
        end_loaded_seq_idx = self.added_data[-1].seq_idx
        if seq_idx < start_loaded_seq_idx or seq_idx > end_loaded_seq_idx:
            raise Exception(
                "seq_idx %i not in loaded seqs range [%i,%i]" % (seq_idx, start_loaded_seq_idx, end_loaded_seq_idx)
            )

    def _get_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq|None
        """
        for data in self.added_data:
            if data.seq_idx == seq_idx:
                return data
        return None

    def is_cached(self, start, end):
        """
        :param int start:
        :param int end:
        :rtype: bool
        """
        # Always False, to force that we call self._load_seqs().
        # This is important for our buffer management.
        return False

    def _load_seqs(self, start, end):
        """
        :param int start: inclusive seq idx start
        :param int end: exclusive seq idx end
        """
        # We expect that start increase monotonic on each call
        # for not-yet-loaded data.
        # This will already be called with _load_seqs_superset indices.
        assert start >= self.expected_load_seq_start
        if start > self.expected_load_seq_start:
            # Cleanup old data.
            self._cleanup_old_seqs(start)
            self.expected_load_seq_start = start
        if self.added_data:
            start = max(self.added_data[-1].seq_idx + 1, start)
        if end > self.num_seqs:
            end = self.num_seqs
        if end >= self.num_seqs:
            self.reached_final_seq = True
        seqs = [self._make_seq(seq_idx) for seq_idx in range(start, end)]
        self._num_timesteps += sum([seq.num_frames for seq in seqs])
        self.added_data += seqs

    def _make_seq(self, seq_idx: int) -> DatasetSeq:
        seq = self.get_corpus_seq(self.get_corpus_seq_idx(seq_idx))
        seq.seq_idx = seq_idx
        return seq

    def have_get_corpus_seq(self) -> bool:
        """
        :return: whether we have :func:`get_corpus_seq`
        """
        return True

    def get_corpus_seq(self, corpus_seq_idx: int) -> DatasetSeq:
        """
        :param corpus_seq_idx:
        :return: seq
        """
        # seed value based on epoch and corpus_seq_idx in order to get deterministic behavior
        self.random.seed((self._get_random_seed_for_epoch(epoch=self.epoch), corpus_seq_idx))
        seq = self.generate_seq(corpus_seq_idx)
        if self.window > 1:
            seq.features["data"] = self._sliding_window(seq.features["data"])
        return seq

    def generate_seq(self, seq_idx: int) -> DatasetSeq:
        """
        This assumes that self.random is already initialized and seeded
        to sth deterministic for the given seq_idx and epoch.

        :param seq_idx: corpus seq idx
        """
        raise NotImplementedError

    def _shuffle_frames_in_seqs(self, start, end):
        assert False, "Shuffling in GeneratingDataset does not make sense."

    def get_num_timesteps(self):
        """
        :rtype: int
        """
        assert self.reached_final_seq
        return self._num_timesteps

    @property
    def num_seqs(self) -> int:
        """
        :return: num seqs for current epoch
        """
        return self._num_seqs

    def get_total_num_seqs(self) -> int:
        """
        :return: total num seqs
        """
        return self._total_num_seqs

    def have_corpus_seq_idx(self):
        """
        :return: whether we have :func:`get_corpus_seq_idx`
        """
        return True

    def get_corpus_seq_idx(self, seq_idx: int) -> int:
        """
        :param seq_idx:
        :return: corpus seq idx
        """
        if self._seq_order is None:
            return seq_idx
        return self._seq_order[seq_idx]

    def get_seq_length(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: returnn.util.NumbersDict
        """
        # get_seq_length() can be called before the seq is loaded via load_seqs().
        # Thus, we just call load_seqs() ourselves here.
        assert seq_idx >= self.expected_load_seq_start
        self.load_seqs(self.expected_load_seq_start, seq_idx + 1)
        return self._get_seq(seq_idx).num_frames

    def get_data(self, seq_idx, key):
        """
        :param int seq_idx:
        :param str key:
        :rtype: numpy.ndarray
        """
        return self._get_seq(seq_idx).features[key]

    def get_input_data(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: numpy.ndarray
        """
        return self.get_data(seq_idx, "data")

    def get_targets(self, target, seq_idx):
        """
        :param int seq_idx:
        :param str target:
        :rtype: numpy.ndarray
        """
        return self.get_data(seq_idx, target)

    def get_tag(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: str
        """
        self._check_loaded_seq_idx(seq_idx)
        return self._get_seq(seq_idx).seq_tag

    def get_all_tags(self):
        """
        :rtype: list[str]
        """
        return ["seq-%i" % seq_idx for seq_idx in range(self.get_total_num_seqs())]

    def get_current_seq_order(self) -> Sequence[int]:
        """
        :return: seq order
        """
        return self._seq_order


class Task12AXDataset(GeneratingDataset):
    """
    12AX memory task.
    This is a simple memory task where there is an outer loop and an inner loop.
    Description here: https://psych.colorado.edu/~oreilly/pubs-abstr.html#OReillyFrank06
    """

    _input_classes = "123ABCXYZ"  # noqa
    _output_classes = "LR"

    def __init__(self, **kwargs):
        super(Task12AXDataset, self).__init__(
            input_dim=len(self._input_classes), output_dim=len(self._output_classes), **kwargs
        )

    def get_random_seq_len(self):
        """
        :rtype: int
        """
        return self.random.randint(10, 100)

    def generate_input_seq(self, seq_len):
        """
        Somewhat made up probability distribution.
        Try to make in a way that at least some "R" will occur in the output seq.
        Otherwise, "R"s are really rare.

        :param int seq_len:
        :rtype: list[int]
        """
        seq = self.random.choice(["", "1", "2"])
        while len(seq) < seq_len:
            if self.random.uniform() < 0.5:
                seq += self.random.choice(list("12"))
            if self.random.uniform() < 0.9:
                seq += self.random.choice(["AX", "BY"])
            while self.random.uniform() < 0.5:
                seq += self.random.choice(list(self._input_classes))
        return list(map(self._input_classes.index, seq[:seq_len]))

    @classmethod
    def make_output_seq(cls, input_seq):
        """
        :type input_seq: list[int]
        :rtype: list[int]
        """
        outer_state = ""
        inner_state = ""
        input_classes = cls._input_classes
        output_seq_str = ""
        for i in input_seq:
            c = input_classes[i]
            o = "L"
            if c in "12":
                outer_state = c
            elif c in "AB":
                inner_state = c
            elif c in "XY":
                if outer_state + inner_state + c in ["1AX", "2BY"]:
                    o = "R"
                inner_state = ""
            # Ignore other cases, "3CZ".
            output_seq_str += o
        return list(map(cls._output_classes.index, output_seq_str))

    def estimate_output_class_priors(self, num_trials, seq_len=10):
        """
        :type num_trials: int
        :param int seq_len:
        :rtype: (float, float)
        """
        count_l, count_r = 0, 0
        for i in range(num_trials):
            input_seq = self.generate_input_seq(seq_len)
            output_seq = self.make_output_seq(input_seq)
            count_l += output_seq.count(0)
            count_r += output_seq.count(1)
        return float(count_l) / (num_trials * seq_len), float(count_r) / (num_trials * seq_len)

    def generate_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        seq_len = self.get_random_seq_len()
        input_seq = self.generate_input_seq(seq_len)
        output_seq = self.make_output_seq(input_seq)
        features = class_idx_seq_to_1_of_k(input_seq, num_classes=len(self._input_classes))
        targets = numpy.array(output_seq, dtype="int32")
        return DatasetSeq(seq_idx=seq_idx, features=features, targets=targets)


class TaskEpisodicCopyDataset(GeneratingDataset):
    """
    Episodic Copy memory task.
    This is a simple memory task where we need to remember a sequence.
    Described in: https://arxiv.org/abs/1511.06464
    Also tested for Associative LSTMs.
    This is a variant where the lengths are random, both for the chars and for blanks.
    """

    # Blank, delimiter and some chars.
    _input_classes = " .01234567"
    _output_classes = _input_classes

    def __init__(self, **kwargs):
        super(TaskEpisodicCopyDataset, self).__init__(
            input_dim=len(self._input_classes), output_dim=len(self._output_classes), **kwargs
        )

    def generate_input_seq(self):
        """
        :rtype: list[int]
        """
        seq = ""
        # Start with random chars.
        rnd_char_len = self.random.randint(1, 10)
        seq += "".join([self.random.choice(list(self._input_classes[2:])) for _ in range(rnd_char_len)])
        blank_len = self.random.randint(1, 100)
        seq += " " * blank_len  # blanks
        seq += "."  # 1 delim
        seq += "." * (rnd_char_len + 1)  # we wait for the outputs + 1 delim
        return list(map(self._input_classes.index, seq))

    @classmethod
    def make_output_seq(cls, input_seq):
        """
        :type input_seq: list[int]
        :rtype: list[int]
        """
        input_classes = cls._input_classes
        input_mem = ""
        output_seq_str = ""
        state = 0
        for i in input_seq:
            c = input_classes[i]
            if state == 0:
                output_seq_str += " "
                if c == " ":
                    pass  # just ignore
                elif c == ".":
                    state = 1  # start with recall now
                else:
                    input_mem += c
            else:  # recall from memory
                # Ignore input.
                if not input_mem:
                    output_seq_str += "."
                else:
                    output_seq_str += input_mem[:1]
                    input_mem = input_mem[1:]
        return list(map(cls._output_classes.index, output_seq_str))

    def generate_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        input_seq = self.generate_input_seq()
        output_seq = self.make_output_seq(input_seq)
        features = class_idx_seq_to_1_of_k(input_seq, num_classes=len(self._input_classes))
        targets = numpy.array(output_seq)
        return DatasetSeq(seq_idx=seq_idx, features=features, targets=targets)


class TaskXmlModelingDataset(GeneratingDataset):
    """
    XML modeling memory task.
    This is a memory task where we need to remember a stack.
    Defined in Jozefowicz et al. (2015).
    Also tested for Associative LSTMs.
    """

    # Blank, XML-tags and some chars.
    _input_classes = " <>/abcdefgh"  # noqa
    _output_classes = _input_classes

    def __init__(self, limit_stack_depth=4, **kwargs):
        super(TaskXmlModelingDataset, self).__init__(
            input_dim=len(self._input_classes), output_dim=len(self._output_classes), **kwargs
        )
        self.limit_stack_depth = limit_stack_depth

    def generate_input_seq(self):
        """
        :rtype: list[int]
        """
        # Because this is a prediction task, start with blank,
        # and the output seq should predict the next char after the blank.
        seq = " "
        xml_stack = []
        while True:
            if not xml_stack or (len(xml_stack) < self.limit_stack_depth and self.random.rand() > 0.6):
                tag_len = self.random.randint(1, 10)
                tag = "".join([self.random.choice(list(self._input_classes[4:])) for _ in range(tag_len)])
                seq += "<%s>" % tag
                xml_stack += [tag]
            else:
                seq += "</%s>" % xml_stack.pop()
            if not xml_stack and self.random.rand() > 0.2:
                break
        return list(map(self._input_classes.index, seq))

    @classmethod
    def make_output_seq(cls, input_seq):
        """
        :type input_seq: list[int]
        :rtype: list[int]
        """
        input_seq_str = "".join(cls._input_classes[i] for i in input_seq)
        xml_stack = []
        output_seq_str = ""
        state = 0
        for c in input_seq_str:
            if c in " >":
                output_seq_str += "<"  # We expect an open char.
                assert state != 1, repr(input_seq_str)
                state = 1  # expect beginning of tag
            elif state == 1:  # in beginning of tag
                output_seq_str += " "  # We don't know yet.
                assert c == "<", repr(input_seq_str)
                state = 2
            elif state == 2:  # first char in tag
                if c == "/":
                    assert xml_stack, repr(input_seq_str)
                    output_seq_str += xml_stack[-1][0]
                    xml_stack[-1] = xml_stack[-1][1:]
                    state = 4  # closing tag
                else:  # opening tag
                    output_seq_str += " "  # We don't know yet.
                    assert c not in " <>/", repr(input_seq_str)
                    state = 3
                    xml_stack += [c]
            elif state == 3:  # opening tag
                output_seq_str += " "  # We don't know.
                xml_stack[-1] += c
            elif state == 4:  # closing tag
                assert xml_stack, repr(input_seq_str)
                if not xml_stack[-1]:
                    output_seq_str += ">"
                    xml_stack.pop()
                    state = 0
                else:
                    output_seq_str += xml_stack[-1][0]
                    xml_stack[-1] = xml_stack[-1][1:]
            else:
                assert False, "invalid state %i. input %r" % (state, input_seq_str)
        return list(map(cls._output_classes.index, output_seq_str))

    def generate_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        input_seq = self.generate_input_seq()
        output_seq = self.make_output_seq(input_seq)
        features = class_idx_seq_to_1_of_k(input_seq, num_classes=len(self._input_classes))
        targets = numpy.array(output_seq)
        return DatasetSeq(seq_idx=seq_idx, features=features, targets=targets)


class TaskVariableAssignmentDataset(GeneratingDataset):
    """
    Variable Assignment memory task.
    This is a memory task to test for key-value retrieval.
    Defined in Associative LSTM paper.
    """

    # Blank/Delim/End, Store/Query, and some chars for key/value.
    _input_classes = " ,.SQ()abcdefgh"  # noqa
    _output_classes = _input_classes

    def __init__(self, **kwargs):
        super(TaskVariableAssignmentDataset, self).__init__(
            input_dim=len(self._input_classes), output_dim=len(self._output_classes), **kwargs
        )

    def generate_input_seq(self):
        """
        :rtype: list[int]
        """
        seq = ""
        from collections import OrderedDict

        store = OrderedDict()
        # First the assignments.
        num_assignments = self.random.randint(1, 5)
        for i in range(num_assignments):
            key_len = self.random.randint(2, 5)
            while True:  # find unique key
                key = "".join([self.random.choice(list(self._input_classes[7:])) for _ in range(key_len)])
                if key not in store:
                    break
            value_len = self.random.randint(1, 2)
            value = "".join([self.random.choice(list(self._input_classes[7:])) for _ in range(value_len)])
            if seq:
                seq += ","
            seq += "S(%s,%s)" % (key, value)
            store[key] = value
        # Now one query.
        key = self.random.choice(store.keys())
        value = store[key]
        seq += ",Q(%s)" % key
        seq += "%s." % value
        return list(map(self._input_classes.index, seq))

    @classmethod
    def make_output_seq(cls, input_seq):
        """
        :type input_seq: list[int]
        :rtype: list[int]
        """
        input_seq_str = "".join(cls._input_classes[i] for i in input_seq)
        store = {}
        key, value = "", ""
        output_seq_str = ""
        state = 0
        for c in input_seq_str:
            if state == 0:
                key = ""
                if c == "S":
                    state = 1  # store
                elif c == "Q":
                    state = 2  # query
                elif c in " ,":
                    pass  # can be ignored
                else:
                    assert False, "c %r in %r" % (c, input_seq_str)
                output_seq_str += " "
            elif state == 1:  # store
                assert c == "(", repr(input_seq_str)
                state = 1.1
                output_seq_str += " "
            elif state == 1.1:  # store.key
                if c == ",":
                    assert key
                    value = ""
                    state = 1.5  # store.value
                else:
                    assert c not in " .,SQ()", repr(input_seq_str)
                    key += c
                output_seq_str += " "
            elif state == 1.5:  # store.value
                if c == ")":
                    assert value
                    store[key] = value
                    state = 0
                else:
                    assert c not in " .,SQ()", repr(input_seq_str)
                    value += c
                output_seq_str += " "
            elif state == 2:  # query
                assert c == "(", repr(input_seq_str)
                state = 2.1
                output_seq_str += " "
            elif state == 2.1:  # query.key
                if c == ")":
                    value = store[key]
                    output_seq_str += value[0]
                    value = value[1:]
                    state = 2.5
                else:
                    assert c not in " .,SQ()", repr(input_seq_str)
                    key += c
                    output_seq_str += " "
            elif state == 2.5:  # query result
                assert c not in " .,SQ()", repr(input_seq_str)
                if value:
                    output_seq_str += value[0]
                    value = value[1:]
                else:
                    output_seq_str += "."
                    state = 2.6
            elif state == 2.6:  # query result end
                assert c == ".", repr(input_seq_str)
                output_seq_str += " "
            else:
                assert False, "invalid state %i, input %r" % (state, input_seq_str)
        return list(map(cls._output_classes.index, output_seq_str))

    def generate_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        input_seq = self.generate_input_seq()
        output_seq = self.make_output_seq(input_seq)
        features = class_idx_seq_to_1_of_k(input_seq, num_classes=len(self._input_classes))
        targets = numpy.array(output_seq)
        return DatasetSeq(seq_idx=seq_idx, features=features, targets=targets)


class TaskNumberBaseConvertDataset(GeneratingDataset):
    """
    Task: E.g: Get some number in octal and convert it to binary (e.g. "10101001").
    Or basically convert some number from some base into another base.
    """

    def __init__(self, input_base=8, output_base=2, min_input_seq_len=1, max_input_seq_len=8, **kwargs):
        """
        :param int input_base:
        :param int output_base:
        :param int min_input_seq_len:
        :param int max_input_seq_len:
        """
        super(TaskNumberBaseConvertDataset, self).__init__(
            input_dim=input_base, output_dim={"data": (input_base, 1), "classes": (output_base, 1)}, **kwargs
        )
        chars = "0123456789abcdefghijklmnopqrstuvwxyz"  # noqa
        assert 2 <= input_base <= len(chars) and 2 <= output_base <= len(chars)
        self.input_base = input_base
        self.output_base = output_base
        self._input_classes = chars[:input_base]
        self._output_classes = chars[:output_base]
        self.labels = {"data": self._input_classes, "classes": self._output_classes}
        assert 0 < min_input_seq_len <= max_input_seq_len
        self.min_input_seq_len = min_input_seq_len
        self.max_input_seq_len = max_input_seq_len

    def get_random_input_seq_len(self):
        """
        :rtype: int
        """
        return self.random.randint(self.min_input_seq_len, self.max_input_seq_len + 1)

    def generate_input_seq(self):
        """
        :rtype: list[int]
        """
        seq_len = self.get_random_input_seq_len()
        seq = [self.random.randint(0, len(self._input_classes)) for _ in range(seq_len)]
        return seq

    def make_output_seq(self, input_seq):
        """
        :param list[int] input_seq:
        :rtype: list[int]
        """
        number = 0
        for i, d in enumerate(reversed(input_seq)):
            number += d * (self.input_base**i)
        output_seq = []
        while number:
            output_seq.insert(0, number % self.output_base)
            number //= self.output_base
        if not output_seq:
            output_seq = [0]
        return output_seq

    def generate_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        input_seq = self.generate_input_seq()
        output_seq = self.make_output_seq(input_seq)
        features = numpy.array(input_seq)
        targets = numpy.array(output_seq)
        return DatasetSeq(seq_idx=seq_idx, features=features, targets=targets)


class DummyDataset(GeneratingDataset):
    """
    Some dummy data, which does not have any meaning.
    If you want to have artificial data with some meaning, look at other datasets here.
    The input are some dense data, the outputs are sparse.
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        num_seqs,
        seq_len=2,
        input_max_value=10.0,
        input_shift=None,
        input_scale=None,
        **kwargs,
    ):
        """
        :param int|None input_dim:
        :param int|dict[str,int|(int,int)|dict] output_dim:
        :param int|float num_seqs:
        :param int|dict[str,int] seq_len:
        :param float input_max_value:
        :param float|None input_shift:
        :param float|None input_scale:
        """
        super(DummyDataset, self).__init__(input_dim=input_dim, output_dim=output_dim, num_seqs=num_seqs, **kwargs)
        self.seq_len = seq_len
        self.input_max_value = input_max_value
        if input_shift is None:
            input_shift = -input_max_value / 2.0
        self.input_shift = input_shift
        if input_scale is None:
            input_scale = 1.0 / self.input_max_value
        self.input_scale = input_scale

    def generate_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        seq_len = self.seq_len
        i1 = seq_idx
        i2 = i1 + seq_len * self.num_inputs
        features = numpy.array(
            [((i % self.input_max_value) + self.input_shift) * self.input_scale for i in range(i1, i2)]
        ).reshape((seq_len, self.num_inputs))
        i1, i2 = i2, i2 + seq_len
        targets = numpy.array([i % self.num_outputs["classes"][0] for i in range(i1, i2)])
        return DatasetSeq(seq_idx=seq_idx, features=features, targets=targets)


class DummyDatasetMultipleSequenceLength(DummyDataset):
    """
    Like :class:`DummyDataset` but has provides seqs with different sequence lengths.
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        num_seqs,
        seq_len=None,
        input_max_value=10.0,
        input_shift=None,
        input_scale=None,
        **kwargs,
    ):
        """
        :param int input_dim:
        :param int output_dim:
        :param int|float num_seqs:
        :param int|dict[str,int] seq_len:
        :param float input_max_value:
        :param float|None input_shift:
        :param float|None input_scale:
        """
        if seq_len is None:
            seq_len = {"data": 10, "classes": 20}
        super(DummyDatasetMultipleSequenceLength, self).__init__(
            input_dim=input_dim,
            output_dim=output_dim,
            num_seqs=num_seqs,
            seq_len=seq_len,
            input_max_value=input_max_value,
            input_shift=input_shift,
            input_scale=input_scale,
            **kwargs,
        )

    def generate_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        assert isinstance(self.seq_len, dict)
        seq_len_data = self.seq_len["data"]
        seq_len_classes = self.seq_len["classes"]
        i1 = seq_idx
        i2 = i1 + seq_len_data * self.num_inputs
        features = numpy.array(
            [((i % self.input_max_value) + self.input_shift) * self.input_scale for i in range(i1, i2)]
        ).reshape((seq_len_data, self.num_inputs))
        i1, i2 = i2, i2 + seq_len_classes
        targets = numpy.array([i % self.num_outputs["classes"][0] for i in range(i1, i2)])
        return DatasetSeq(seq_idx=seq_idx, features=features, targets=targets)


class DummyDatasetMultipleDataKeys(DummyDataset):
    """
    Like :class:`DummyDataset` this class provides dummy data without any meaning.
    But it extends :class:`DummyDataset` such that it is able to provide data for multiple data keys,
    not only `"data"` and `"classes"` (those are also overridable, though the current implementation
    expects a `"data"` key).
    Further, `output_dim` is expected to be a `dict` now, which defines the data format for each
    data key, which also enables the user to customize whether the data is sparse or dense.
    It also provides the function of :class:`DummyDatasetMultipleSequenceLength` to customize the
    sequence length for each data point.
    """

    def __init__(
        self,
        output_dim,
        num_seqs,
        seq_len=None,
        input_max_value=10.0,
        input_shift=None,
        input_scale=None,
        data_keys=None,
        **kwargs,
    ):
        """
        :param dict[str,int|(int,int)|dict] output_dim: `dict` defining the output for each data key
          (e.g. `{"data": [200, 2], "classes": [100, 1]}`).
        :param int|float num_seqs:
        :param int|dict[str,int] seq_len: definition of the sequence length for each data key,
          if `int` the given length is used for all data keys.
        :param float input_max_value:
        :param float|None input_shift:
        :param float|None input_scale:
        :param list[str]|None data_keys: explicit declaration of the data keys,
          if `None` `"data"` and `"classes"` are used.
        """
        if data_keys is None:
            data_keys = ["data", "classes"]
        self.data_keys = data_keys

        _seq_len = 20
        if isinstance(seq_len, int):
            _seq_len = seq_len
            seq_len = None
        if seq_len is None:
            seq_len = {}
            for key in self.data_keys:
                seq_len[key] = _seq_len
        assert set(data_keys) == set(
            seq_len.keys()
        ), "%s: the keys of seq_len (%s) must match the keys in data_keys=%s." % (
            self,
            str(seq_len.keys()),
            str(data_keys),
        )
        assert isinstance(
            output_dim, dict
        ), "%s: output_dim %r must be a dict containing a definition for each key in data_keys." % (self, output_dim)
        assert set(data_keys) == set(
            output_dim.keys()
        ), "%s: the keys of output_dim (%s) must match the keys in data_keys=%s." % (
            self,
            str(output_dim.keys()),
            str(data_keys),
        )

        super(DummyDatasetMultipleDataKeys, self).__init__(
            input_dim=None,  # this was only used for the definition of "data", but this is handled by `output_dim` now.
            output_dim=output_dim,
            num_seqs=num_seqs,
            seq_len=seq_len,
            input_max_value=input_max_value,
            input_shift=input_shift,
            input_scale=input_scale,
            **kwargs,
        )

    def generate_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        features = {}
        i1 = seq_idx

        for key in self.data_keys:
            seq_len = self.seq_len[key]
            output_dim = self.num_outputs[key][0]
            is_sparse = self.num_outputs[key][1] == 1

            if is_sparse:
                i2 = i1 + seq_len
                features[key] = numpy.array([i % self.num_outputs[key][0] for i in range(i1, i2)])
            else:
                i2 = i1 + seq_len * output_dim
                features[key] = numpy.array(
                    [((i % self.input_max_value) + self.input_shift) * self.input_scale for i in range(i1, i2)]
                ).reshape((seq_len, output_dim))
            i1 = i2

        return DatasetSeq(seq_idx=seq_idx, features=features, targets=None)


class StaticDataset(CachedDataset2):
    """
    Provide all the data as a list of dict of numpy arrays.
    """

    @classmethod
    def copy_from_dataset(cls, dataset, start_seq_idx=0, max_seqs=None):
        """
        :param Dataset dataset:
        :param int start_seq_idx:
        :param int|None max_seqs:
        :rtype: StaticDataset
        """
        if isinstance(dataset, StaticDataset):
            return cls(
                data=dataset.data,
                target_list=dataset.target_list,
                output_dim=dataset.num_outputs,
                input_dim=dataset.num_inputs,
            )
        seq_idx = start_seq_idx
        data = []
        while dataset.is_less_than_num_seqs(seq_idx):
            dataset.load_seqs(seq_idx, seq_idx + 1)
            if max_seqs is not None and len(data) >= max_seqs:
                break
            seq_data = {key: dataset.get_data(seq_idx, key) for key in dataset.get_data_keys()}
            data.append(seq_data)
            seq_idx += 1
        return cls(
            data=data,
            target_list=dataset.get_target_list(),
            output_dim=dataset.num_outputs,
            input_dim=dataset.num_inputs,
        )

    def __init__(self, data, target_list=None, input_dim=None, output_dim=None, **kwargs):
        """
        :param list[dict[str,numpy.ndarray]] data: list of seqs, each provide the data for each data-key
        :param target_list:
        :param int|None input_dim:
        :param int|dict[str,(int,int)|list[int]] output_dim:
        """
        super(StaticDataset, self).__init__(**kwargs)

        assert len(data) > 0
        self.data = data
        first_data = data[0]
        self.data_keys = sorted(first_data.keys())
        if "data" in self.data_keys:
            self.data_keys.remove("data")
            self.data_keys.insert(0, "data")
        if target_list is not None:
            for key in target_list:
                assert key in self.data_keys
        else:
            target_list = list(self.data_keys)
            if "data" in target_list:
                target_list.remove("data")
        self.target_list = target_list

        if output_dim is None:
            output_dim = {}
        output_dim = convert_data_dims(output_dim, leave_dict_as_is=False)
        if input_dim is not None and "data" not in output_dim:
            assert "data" in self.data_keys
            output_dim["data"] = (input_dim, 2)  # assume dense, not sparse
        for key, value in first_data.items():
            if key not in output_dim:
                output_dim[key] = (value.shape[-1] if value.ndim >= 2 else 0, len(value.shape))
        if input_dim is None and "data" in self.data_keys:
            input_dim = output_dim["data"][0]
        for key in self.data_keys:
            first_data_output = first_data[key]
            assert key in output_dim
            assert output_dim[key][1] == len(first_data_output.shape)
            if len(first_data_output.shape) >= 2:
                assert output_dim[key][0] == first_data_output.shape[-1]
        assert set(output_dim.keys()) == set(self.data_keys), "output_dim does not match the given data"

        self.num_inputs = input_dim
        output_dim = convert_data_dims(output_dim, leave_dict_as_is=False)
        if "data" not in output_dim and input_dim is not None:
            output_dim["data"] = (input_dim * self.window, 2)  # not sparse
        self.num_outputs = output_dim

        self._seq_order = None
        self.init_seq_order(epoch=1)

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :param int|None epoch:
        :param list[str]|None seq_list: List of sequence tags, to set a predefined order.
        :param list[int]|None seq_order: List of corpus sequence indices, to set a predefined order. Only possible
          if the dataset has such indices (see self.have_corpus_seq_idx()).
        :rtype: bool
        :returns whether the order changed (True is always safe to return)
        """
        super(StaticDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        if seq_order is not None:
            assert not seq_list
            self._seq_order = seq_order
        elif seq_list is not None:
            import re

            # If the re.match fails here, some seq tag is invalid.
            self._seq_order = [int(re.match("^seq-(\\d+)$", seq).group(1)) for seq in seq_list]
        else:
            default_data_key = self.data_keys[0]
            self._seq_order = self.get_seq_order_for_epoch(
                epoch=epoch, num_seqs=len(self.data), get_seq_len=lambda i: self.data[i][default_data_key].shape[0]
            )
        self._num_seqs = len(self._seq_order)
        return True

    def _collect_single_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        corpus_seq_idx = self._seq_order[seq_idx]
        data = self.data[corpus_seq_idx]
        return DatasetSeq(
            seq_idx=seq_idx, seq_tag="seq-%i" % corpus_seq_idx, features={key: data[key] for key in self.data_keys}
        )

    def get_data_keys(self):
        """
        :rtype: list[str]
        """
        return self.data_keys

    def get_target_list(self):
        """
        :rtype: list[str]
        """
        return self.target_list

    def get_data_dtype(self, key):
        """
        :param str key:
        :rtype: str
        """
        return self.data[0][key].dtype

    def get_total_num_seqs(self):
        """
        :rtype: int
        """
        return len(self.data)

    def get_all_tags(self):
        """
        :return: list of all seq tags, of the whole dataset, without partition epoch.
        :rtype: list[str]
        """
        return ["seq-%i" % i for i in range(self.get_total_num_seqs())]

    def get_tag(self, sorted_seq_idx):
        """
        :param int sorted_seq_idx:
        :rtype: str
        """
        return "seq-%i" % self.get_corpus_seq_idx(sorted_seq_idx)

    def have_corpus_seq_idx(self):
        """
        :rtype: bool
        :return: whether you can call self.get_corpus_seq_idx()
        """
        return True

    def get_corpus_seq_idx(self, seq_idx):
        """
        :param int seq_idx: sorted sequence index from the current epoch, depending on seq_ordering
        :return: the sequence index as-is in the original corpus (as if you would have sorting="default").
        :rtype: int
        """
        return self._seq_order[seq_idx]


class CopyTaskDataset(GeneratingDataset):
    """
    Copy task.
    Input/output is exactly the same random sequence of sparse labels.
    """

    def __init__(self, nsymbols, minlen=0, maxlen=0, minlen_epoch_factor=0, maxlen_epoch_factor=0, **kwargs):
        """
        :param int nsymbols:
        :param int minlen:
        :param int maxlen:
        :param float minlen_epoch_factor:
        :param float maxlen_epoch_factor:
        """
        # Sparse data.
        super(CopyTaskDataset, self).__init__(
            input_dim=nsymbols, output_dim={"data": (nsymbols, 1), "classes": (nsymbols, 1)}, **kwargs
        )

        assert nsymbols <= 256
        self.nsymbols = nsymbols
        self.minlen = minlen
        self.maxlen = maxlen
        self.minlen_epoch_factor = minlen_epoch_factor
        self.maxlen_epoch_factor = maxlen_epoch_factor

    def get_random_seq_len(self):
        """
        :rtype: int
        """
        assert isinstance(self.epoch, int)
        minlen = int(self.minlen + self.minlen_epoch_factor * self.epoch)
        maxlen = int(self.maxlen + self.maxlen_epoch_factor * self.epoch)
        assert 0 < minlen <= maxlen
        return self.random.randint(minlen, maxlen + 1)

    def generate_seq(self, seq_idx):
        """
        :type seq_idx: int
        :rtype: DatasetSeq
        """
        seq_len = self.get_random_seq_len()
        seq = [self.random.randint(0, self.nsymbols) for _ in range(seq_len)]
        seq_np = numpy.array(seq, dtype="int8")
        return DatasetSeq(seq_idx=seq_idx, features=seq_np, targets={"classes": seq_np})


# Multiple external sources where we could write automatic wrappers:
# * https://github.com/tensorflow/datasets
# * tf.contrib.keras.datasets, https://www.tensorflow.org/api_docs/python/tf/keras/datasets
# * nltk.corpus


class TimitDataset(CachedDataset2):
    """
    DARPA TIMIT Acoustic-Phonetic Continuous Speech Corpus.
    You must provide the data.

    Demo:

        tools/dump-dataset.py "{'class': 'TimitDataset', 'timit_dir': '...'}"
        tools/dump-dataset.py "{'class': 'TimitDataset', 'timit_dir': '...',
                                'demo_play_audio': True, 'random_permute_audio': True}"

    The full train data has 3696 utterances and the core test data has 192 utterances
    (24-speaker core test set).

    The input length is not the same as the output length.
    The targets are not the framewise alignment.

    For some references:
    https://github.com/ppwwyyxx/tensorpack/blob/master/examples/CTC-TIMIT/train-timit.py
    https://www.cs.toronto.edu/~graves/preprint.pdf
    https://arxiv.org/pdf/1303.5778.pdf
    https://arxiv.org/pdf/0804.3269.pdf
    """

    # via: https://github.com/kaldi-asr/kaldi/blob/master/egs/timit/s5/conf/phones.60-48-39.map
    PhoneMapTo39 = {
        "aa": "aa",
        "ae": "ae",
        "ah": "ah",
        "ao": "aa",
        "aw": "aw",
        "ax": "ah",
        "ax-h": "ah",
        "axr": "er",
        "ay": "ay",
        "b": "b",
        "bcl": "sil",
        "ch": "ch",
        "d": "d",
        "dcl": "sil",
        "dh": "dh",
        "dx": "dx",
        "eh": "eh",
        "el": "l",
        "em": "m",
        "en": "n",
        "eng": "ng",
        "epi": "sil",
        "er": "er",
        "ey": "ey",
        "f": "f",
        "g": "g",
        "gcl": "sil",
        "h#": "sil",
        "hh": "hh",
        "hv": "hh",
        "ih": "ih",
        "ix": "ih",
        "iy": "iy",
        "jh": "jh",
        "k": "k",
        "kcl": "sil",
        "l": "l",
        "m": "m",
        "n": "n",
        "ng": "ng",
        "nx": "n",
        "ow": "ow",
        "oy": "oy",
        "p": "p",
        "pau": "sil",
        "pcl": "sil",
        "q": None,
        "r": "r",
        "s": "s",
        "sh": "sh",
        "t": "t",
        "tcl": "sil",
        "th": "th",
        "uh": "uh",
        "uw": "uw",
        "ux": "uw",
        "v": "v",
        "w": "w",
        "y": "y",
        "z": "z",
        "zh": "sh",
    }
    PhoneMapTo48 = {
        "aa": "aa",
        "ae": "ae",
        "ah": "ah",
        "ao": "ao",
        "aw": "aw",
        "ax": "ax",
        "ax-h": "ax",
        "axr": "er",
        "ay": "ay",
        "b": "b",
        "bcl": "vcl",
        "ch": "ch",
        "d": "d",
        "dcl": "vcl",
        "dh": "dh",
        "dx": "dx",
        "eh": "eh",
        "el": "el",
        "em": "m",
        "en": "en",
        "eng": "ng",
        "epi": "epi",
        "er": "er",
        "ey": "ey",
        "f": "f",
        "g": "g",
        "gcl": "vcl",
        "h#": "sil",
        "hh": "hh",
        "hv": "hh",
        "ih": "ih",
        "ix": "ix",
        "iy": "iy",
        "jh": "jh",
        "k": "k",
        "kcl": "cl",
        "l": "l",
        "m": "m",
        "n": "n",
        "ng": "ng",
        "nx": "n",
        "ow": "ow",
        "oy": "oy",
        "p": "p",
        "pau": "sil",
        "pcl": "cl",
        "q": None,
        "r": "r",
        "s": "s",
        "sh": "sh",
        "t": "t",
        "tcl": "cl",
        "th": "th",
        "uh": "uh",
        "uw": "uw",
        "ux": "uw",
        "v": "v",
        "w": "w",
        "y": "y",
        "z": "z",
        "zh": "zh",
    }
    Phones61 = PhoneMapTo39.keys()
    PhoneMapTo61 = {p: p for p in Phones61}

    @classmethod
    def _get_phone_map(cls, num_phones=61):
        """
        :param int num_phones:
        :return: map 61-phone-set-phone -> num_phones-phone-set-phone
        :rtype: dict[str,str|None]
        """
        return {61: cls.PhoneMapTo61, 48: cls.PhoneMapTo48, 39: cls.PhoneMapTo39}[num_phones]

    @classmethod
    def _get_labels_by_phone_map(cls, phone_map):
        """
        :param dict[str,str|None] phone_map:
        :rtype: list[str]
        """
        labels = sorted(set(filter(None, phone_map.values())))
        # Make 'sil' the 0 phoneme.
        if "pau" in labels:
            labels.remove("pau")
            labels.insert(0, "pau")
        else:
            labels.remove("sil")
            labels.insert(0, "sil")
        return labels

    @classmethod
    def get_labels(cls, num_phones=61):
        """
        :param int num_phones:
        :rtype: list[str]
        """
        return cls._get_labels_by_phone_map(cls._get_phone_map(num_phones))

    @classmethod
    def get_label_map(cls, source_num_phones=61, target_num_phones=39):
        """
        :param int source_num_phones:
        :param int target_num_phones:
        :rtype: dict[int,int|None]
        """
        src_phone_map = cls._get_phone_map(source_num_phones)  # 61-phone -> src-phone
        src_labels = cls._get_labels_by_phone_map(src_phone_map)  # src-idx -> src-phone
        tgt_phone_map = cls._get_phone_map(target_num_phones)  # 61-phone -> tgt-phone
        tgt_labels = cls._get_labels_by_phone_map(tgt_phone_map)  # tgt-idx -> tgt-phone
        d = {i: src_labels[i] for i in range(source_num_phones)}  # src-idx -> src-phone|61-phone
        if source_num_phones != 61:
            src_phone_map_rev = {v: k for (k, v) in sorted(src_phone_map.items())}  # src-phone -> 61-phone
            d = {i: src_phone_map_rev[v] for (i, v) in d.items()}  # src-idx -> 61-phone
        d = {i: tgt_phone_map[v] for (i, v) in d.items()}  # src-idx -> tgt-phone
        d = {i: tgt_labels.index(v) if v else None for (i, v) in d.items()}  # src-idx -> tgt-idx
        return d

    def __init__(
        self,
        timit_dir,
        train=True,
        preload=False,
        features="mfcc",
        num_feature_filters=40,
        feature_window_len=0.025,
        feature_step_len=0.010,
        with_delta=False,
        norm_mean=None,
        norm_std_dev=None,
        random_permute_audio=None,
        num_phones=61,
        demo_play_audio=False,
        **kwargs,
    ):
        """
        :param str|None timit_dir: directory of TIMIT. should contain train/filelist.phn and test/filelist.core.phn
        :param bool train: whether to use the train or core test data
        :param bool preload: if True, here at __init__, we will wait until we loaded all the data
        :param str|function features: see :class:`ExtractAudioFeatures`
        :param int num_feature_filters: e.g. number of MFCCs
        :param bool|int with_delta: whether to add delta features (doubles the features dim). if int, up to this degree
        :param str norm_mean: file with mean values which are used for mean-normalization of the final features
        :param str norm_std_dev: file with std dev valeus for variance-normalization of the final features
        :param None|bool|dict[str] random_permute_audio: enables permutation on the audio.
            see _get_random_permuted_audio
        :param int num_phones: 39, 48 or 61. num labels of our classes
        :param bool demo_play_audio: plays the audio. only make sense with tools/dump-dataset.py
        """
        super(TimitDataset, self).__init__(**kwargs)
        from threading import Lock, Thread

        self._lock = Lock()
        self._features = features
        self._num_feature_filters = num_feature_filters
        self._feature_window_len = feature_window_len
        self._feature_step_len = feature_step_len
        self.num_inputs = self._num_feature_filters
        if isinstance(with_delta, bool):
            with_delta = 1 if with_delta else 0
        assert isinstance(with_delta, int) and with_delta >= 0
        self._with_delta = with_delta
        self.num_inputs *= 1 + with_delta
        self._norm_mean = self._load_feature_vec(norm_mean)
        self._norm_std_dev = self._load_feature_vec(norm_std_dev)
        assert num_phones in {61, 48, 39}
        self._phone_map = {61: self.PhoneMapTo61, 48: self.PhoneMapTo48, 39: self.PhoneMapTo39}[num_phones]
        self.labels = self._get_labels_by_phone_map(self._phone_map)
        self.num_outputs = {"data": (self.num_inputs, 2), "classes": (len(self.labels), 1)}
        self._timit_dir = timit_dir
        self._is_train = train
        self._demo_play_audio = demo_play_audio
        self._random = numpy.random.RandomState(1)
        if random_permute_audio is None:
            random_permute_audio = train
        from returnn.util.basic import CollectionReadCheckCovered

        self._random_permute_audio = CollectionReadCheckCovered.from_bool_or_dict(random_permute_audio)

        self._seq_order = None  # type: typing.Optional[typing.Sequence[int]]
        self._init_timit()

        self._audio_data = {}  # seq_tag -> (audio, sample_rate). loaded by self._reader_thread_main
        self._phone_seqs = {}  # seq_tag -> phone_seq (list of str)
        self._reader_thread = Thread(name="%r reader" % self, target=self._reader_thread_main)
        self._reader_thread.daemon = True
        self._reader_thread.start()
        if preload:
            self._preload()

    def _load_feature_vec(self, value):
        """
        :param str|None value:
        :return: shape (self.num_inputs,), float32
        :rtype: numpy.ndarray|None
        """
        if value is None:
            return None
        if isinstance(value, str):
            value = numpy.loadtxt(value)
        assert isinstance(value, numpy.ndarray)
        assert value.shape == (self.num_inputs,)
        return value.astype("float32")

    def _init_timit(self):
        """
        Sets self._seq_tags, _num_seqs, _seq_order, and _timit_dir.
        timit_dir should be such that audio_filename = "%s/%s.wav" % (timit_dir, seq_tag).
        """
        import os

        assert os.path.exists(self._timit_dir)
        if self._is_train:
            self._timit_dir += "/train"
        else:
            self._timit_dir += "/test"
        assert os.path.exists(self._timit_dir)
        if self._is_train:
            file_list_fn = self._timit_dir + "/filelist.phn"
        else:
            file_list_fn = self._timit_dir + "/filelist.core.phn"
        assert os.path.exists(file_list_fn)
        seq_tags = [os.path.splitext(p)[0] for p in open(file_list_fn).read().splitlines()]
        self._seq_tags = seq_tags
        self._num_seqs = len(self._seq_tags)
        self._seq_order = list(range(self._num_seqs))

    def _preload(self):
        import time

        last_print_time = 0
        last_print_len = None
        while True:
            with self._lock:
                cur_len = len(self._audio_data)
            if cur_len == len(self._seq_tags):
                return
            if cur_len != last_print_len and time.time() - last_print_time > 10:
                print("%r: loading (%i/%i loaded so far)..." % (self, cur_len, len(self._seq_tags)), file=log.v3)
                last_print_len = cur_len
                last_print_time = time.time()
            time.sleep(1)

    def _reader_thread_main(self):
        import sys
        from returnn.util.basic import interrupt_main

        # noinspection PyBroadException
        try:
            from returnn.util import better_exchook

            better_exchook.install()

            import librosa  # noqa

            for seq_tag in self._seq_tags:
                audio_filename = "%s/%s.wav" % (self._timit_dir, seq_tag)
                # Don't use librosa.load which internally uses audioread which would use Gstreamer as a backend,
                # which has multiple issues:
                # https://github.com/beetbox/audioread/issues/62
                # https://github.com/beetbox/audioread/issues/63
                # Instead, use PySoundFile, which is also faster. See here for discussions:
                # https://github.com/beetbox/audioread/issues/64
                # https://github.com/librosa/librosa/issues/681
                import soundfile  # noqa  # pip install pysoundfile

                audio, sample_rate = soundfile.read(audio_filename)
                with self._lock:
                    self._audio_data[seq_tag] = (audio, sample_rate)
                phone_seq = self._read_phone_seq(seq_tag)
                with self._lock:
                    self._phone_seqs[seq_tag] = phone_seq

        except Exception:
            sys.excepthook(*sys.exc_info())
            interrupt_main()

    def _read_phn_file(self, seq_tag):
        """
        :param str seq_tag:
        :rtype: list[str]
        """
        import os

        phn_fn = "%s/%s.phn" % (self._timit_dir, seq_tag)
        assert os.path.exists(phn_fn)
        phone_seq = []
        for line in open(phn_fn).read().splitlines():
            t0, t1, p = line.split()
            phone_seq.append(p)
        return phone_seq

    def _read_phone_seq(self, seq_tag):
        """
        :param str seq_tag: e.g. "dr1-fvmh0/s1" or "dr1/fcjf0/sa1"
        :rtype: list[str]
        """
        return self._read_phn_file(seq_tag)

    def _get_phone_seq(self, seq_tag):
        """
        :param str seq_tag: e.g. "dr1-fvmh0/s1" or "dr1/fcjf0/sa1"
        :rtype: list[str]
        """
        import time

        last_print_time = 0
        last_print_len = None
        idx = None
        while True:
            with self._lock:
                if seq_tag in self._phone_seqs:
                    return self._phone_seqs[seq_tag]
                cur_len = len(self._phone_seqs)
            if idx is None:
                idx = self._seq_tags.index(seq_tag)
            if cur_len != last_print_len and time.time() - last_print_time > 10:
                print(
                    "%r: waiting for %r, idx %i (%i/%i loaded so far)..."
                    % (self, seq_tag, idx, cur_len, len(self._seq_tags)),
                    file=log.v3,
                )
                last_print_len = cur_len
                last_print_time = time.time()
            time.sleep(1)

    def _get_audio(self, seq_tag):
        """
        :param str seq_tag: e.g. "dr1-fvmh0/s1" or "dr1/fcjf0/sa1"
        :return: audio, sample_rate
        :rtype: (numpy.ndarray, int)
        """
        import time

        last_print_time = 0
        last_print_len = None
        idx = None
        while True:
            with self._lock:
                if seq_tag in self._audio_data:
                    return self._audio_data[seq_tag]
                cur_len = len(self._audio_data)
            if idx is None:
                idx = self._seq_tags.index(seq_tag)
            if cur_len != last_print_len and time.time() - last_print_time > 10:
                print(
                    "%r: waiting for %r, idx %i (%i/%i loaded so far)..."
                    % (self, seq_tag, idx, cur_len, len(self._seq_tags)),
                    file=log.v3,
                )
                last_print_len = cur_len
                last_print_time = time.time()
            time.sleep(1)

    # noinspection PyMethodMayBeStatic
    def _demo_audio_play(self, audio, sample_rate):
        """
        :param numpy.ndarray audio: shape (sample_len,)
        :param int sample_rate:
        """
        assert audio.dtype == numpy.float32
        assert audio.ndim == 1
        try:
            # noinspection PyPackageRequirements
            import pyaudio
        except ImportError:
            print("pip3 install --user pyaudio")
            raise
        p = pyaudio.PyAudio()
        chunk_size = 1024
        stream = p.open(
            format=pyaudio.paFloat32, channels=1, rate=sample_rate, frames_per_buffer=chunk_size, output=True
        )
        while len(audio) > 0:
            chunk = audio[:chunk_size]
            audio = audio[chunk_size:]
            stream.write(chunk, num_frames=len(chunk))
        stream.stop_stream()
        stream.close()
        p.terminate()

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :param int epoch:
        :param list[str]|None seq_list:
        :param list[int]|None seq_order:
        :rtype: bool
        """
        assert seq_list is None and seq_order is None
        super(TimitDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        self._num_seqs = len(self._seq_tags)
        self._seq_order = self.get_seq_order_for_epoch(
            epoch=epoch, num_seqs=self._num_seqs, get_seq_len=lambda i: len(self._seq_tags[i][1])
        )
        self._random.seed(self._get_random_seed_for_epoch(epoch=epoch))
        return True

    def _collect_single_seq(self, seq_idx):
        """
        :type seq_idx: int
        :rtype: DatasetSeq | None
        :returns DatasetSeq or None if seq_idx >= num_seqs.
        """
        if seq_idx >= len(self._seq_order):
            return None

        seq_tag = self._seq_tags[self._seq_order[seq_idx]]
        phone_seq = self._get_phone_seq(seq_tag)
        phone_seq = [self._phone_map[p] for p in phone_seq]
        phone_seq = [p for p in phone_seq if p]
        phone_id_seq = numpy.array([self.labels.index(p) for p in phone_seq], dtype="int32")
        # see: https://github.com/rdadolf/fathom/blob/master/fathom/speech/preproc.py
        # and: https://groups.google.com/forum/#!topic/librosa/V4Z1HpTKn8Q
        audio, sample_rate = self._get_audio(seq_tag)
        audio_feature_extractor = ExtractAudioFeatures(
            features=self._features,
            num_feature_filters=self._num_feature_filters,
            with_delta=self._with_delta,
            window_len=self._feature_window_len,
            step_len=self._feature_step_len,
            norm_mean=self._norm_mean,
            norm_std_dev=self._norm_std_dev,
            random_permute=self._random_permute_audio,
            random_state=self._random,
        )
        mfccs = audio_feature_extractor.get_audio_features(audio=audio, sample_rate=sample_rate, seq_name=seq_tag)
        return DatasetSeq(seq_idx=seq_idx, seq_tag=seq_tag, features=mfccs, targets=phone_id_seq)


class NltkTimitDataset(TimitDataset):
    """
    DARPA TIMIT Acoustic-Phonetic Continuous Speech Corpus

    This Dataset will get TIMIT via NLTK.
    Demo:

        tools/dump-dataset.py "{'class': 'NltkTimitDataset'}"
        tools/dump-dataset.py "{'class': 'NltkTimitDataset', 'demo_play_audio': True, 'random_permute_audio': True}"

    Note: The NLTK data only contains a subset of the train data (160 utterances),
    and none of the test data.
    The full train data has 3696 utterances and the core test data has 192 utterances.
    Not sure how useful this is...

    See :class:`TimitDataset` for more.
    """

    def __init__(self, nltk_download_dir=None, **kwargs):
        self._nltk_download_dir = nltk_download_dir
        super(NltkTimitDataset, self).__init__(timit_dir=None, **kwargs)

    # noinspection PyPackageRequirements
    def _init_timit(self):
        """
        Sets self._seq_tags, _num_seqs, _seq_order, and _timit_dir.
        timit_dir should be such that audio_filename = "%s/%s.wav" % (timit_dir, seq_tag).
        """
        import os
        from nltk.downloader import Downloader  # noqa

        downloader = Downloader(download_dir=self._nltk_download_dir)
        print("NLTK corpus download dir:", downloader.download_dir, file=log.v3)
        timit_dir = downloader.download_dir + "/corpora/timit"
        if not os.path.exists(timit_dir):
            assert downloader.download("timit")
            assert os.path.exists(timit_dir)
        assert os.path.exists(timit_dir + "/timitdic.txt"), "TIMIT download broken? remove the directory %r" % timit_dir
        self._timit_dir = timit_dir

        from nltk.data import FileSystemPathPointer  # noqa
        from nltk.corpus.reader.timit import TimitCorpusReader  # noqa

        self._data_reader = TimitCorpusReader(FileSystemPathPointer(timit_dir))
        utterance_ids = self._data_reader.utteranceids()
        assert isinstance(utterance_ids, list)
        assert utterance_ids

        # NLTK only has this single set, thus split it into train/dev.
        split = int(len(utterance_ids) * 0.9)
        if self._is_train:
            utterance_ids = utterance_ids[:split]
        else:
            utterance_ids = utterance_ids[split:]
        self._seq_tags = utterance_ids  # list of seq_tag

        self._num_seqs = len(self._seq_tags)
        self._seq_order = list(range(self._num_seqs))

    def _read_phone_seq(self, seq_tag):
        return self._data_reader.phones(seq_tag)


class BlissDataset(CachedDataset2):
    """
    Reads in a Bliss XML corpus (similar to :class:`LmDataset`),
    and provides the features (similar to :class:`TimitDataset`)
    and the orthography as words, subwords or chars (similar to :class:`TranslationDataset`).

    Example:
      ./tools/dump-dataset.py "
        {'class':'BlissDataset',
         'path': '/u/tuske/work/ASR/switchboard/corpus/xml/train.corpus.gz',
         'bpe_file': '/u/zeyer/setups/switchboard/subwords/swb-bpe-codes',
         'vocab_file': '/u/zeyer/setups/switchboard/subwords/swb-vocab'}"
    """

    class SeqInfo:
        """
        Covers all relevant seq info.
        """

        __slots__ = ("idx", "tag", "orth_raw", "orth_seq", "audio_path", "audio_start", "audio_end")

    def __init__(
        self,
        path,
        vocab_file,
        bpe_file=None,
        num_feature_filters=40,
        feature_window_len=0.025,
        feature_step_len=0.010,
        with_delta=False,
        norm_mean=None,
        norm_std_dev=None,
        **kwargs,
    ):
        """
        :param str path: path to XML. can also be gzipped.
        :param str vocab_file: path to vocabulary file. Python-str which evals to dict[str,int]
        :param str bpe_file: Byte-pair encoding file
        :param int num_feature_filters: e.g. number of MFCCs
        :param bool|int with_delta: whether to add delta features (doubles the features dim). if int, up to this degree
        """
        super(BlissDataset, self).__init__(**kwargs)
        assert norm_mean is None and norm_std_dev is None, "%s, not yet implemented..." % self
        from returnn.util.basic import hms_fraction
        import time

        start_time = time.time()
        self._num_feature_filters = num_feature_filters
        self.num_inputs = num_feature_filters
        self._feature_window_len = feature_window_len
        self._feature_step_len = feature_step_len
        if isinstance(with_delta, bool):
            with_delta = 1 if with_delta else 0
        assert isinstance(with_delta, int)
        self._with_delta = with_delta
        self.num_inputs *= 1 + with_delta
        self._bpe_file = open(bpe_file, "r")
        self._seqs = []  # type: typing.List[BlissDataset.SeqInfo]
        self._vocab = {}  # type: typing.Dict[str,int]  # set in self._parse_vocab
        self._parse_bliss_xml(filename=path)
        # TODO: loading audio like in TimitDataset, and in parallel
        self._bpe = BytePairEncoding(vocab_file=vocab_file, bpe_file=bpe_file)
        self.labels["classes"] = self._bpe.labels
        self.num_outputs = {"data": (self.num_inputs, 2), "classes": (self._bpe.num_labels, 1)}
        print(
            "%s: Loaded %r, num seqs: %i, elapsed: %s"
            % (self.__class__.__name__, path, len(self._seqs), hms_fraction(time.time() - start_time)),
            file=log.v3,
        )

    def _parse_bliss_xml(self, filename):
        """
        This takes e.g. around 5 seconds for the Switchboard 300h train corpus.
        Should be as fast as possible to get a list of the segments.
        All further parsing and loading can then be done in parallel and lazily.
        :param str filename:
        :return: nothing, fills self._segments
        """
        # Also see LmDataset._iter_bliss.
        import gzip
        import xml.etree.ElementTree as ElementTree

        corpus_file = open(filename, "rb")
        if filename.endswith(".gz"):
            corpus_file = gzip.GzipFile(fileobj=corpus_file)
        SeqInfo = self.SeqInfo  # noqa
        context = iter(ElementTree.iterparse(corpus_file, events=("start", "end")))
        elem_tree = []
        name_tree = []
        cur_recording = None
        idx = 0
        for event, elem in context:
            if event == "start":
                elem_tree += [elem]
                name_tree += [elem.attrib.get("name", None)]
                if elem.tag == "recording":
                    cur_recording = elem.attrib["audio"]
            if event == "end" and elem.tag == "segment":
                info = SeqInfo()
                info.idx = idx
                info.tag = "/".join(name_tree)
                info.orth_raw = elem.find("orth").text or ""
                info.audio_path = cur_recording
                info.audio_start = float(elem.attrib["start"])
                info.audio_end = float(elem.attrib["end"])
                self._seqs.append(info)
                idx += 1
                if elem_tree:
                    elem_tree[0].clear()  # free memory
            if event == "end":
                assert elem_tree[-1] is elem
                elem_tree = elem_tree[:-1]
                name_tree = name_tree[:-1]
        self._num_seqs = len(self._seqs)

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :param int|None epoch:
        :param list[str]|None seq_list: Predefined order via list of tags, not used here.
        :param list[int]|None seq_order: Predefined order via list of indices, not used here.
        :rtype: bool
        :returns whether the order changed (True is always safe to return)
        """
        super(BlissDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        self._num_seqs = len(self._seqs)
        return True

    def _collect_single_seq(self, seq_idx):
        raise NotImplementedError  # TODO...


class LibriSpeechCorpus(CachedDataset2):
    """
    LibriSpeech. https://www.openslr.org/12/

    "train-*" Seq-length 'data' Stats (default MFCC, every 10ms):
      281241 seqs
      Mean: 1230.94154835176
      Std dev: 383.5126785278322
      Min/max: 84 / 2974
    "train-*" Seq-length 'classes' Stats (BPE with 10k symbols):
      281241 seqs
      Mean: 58.46585312952222
      Std dev: 20.54464373013634
      Min/max: 1 / 161
    "train-*" mean transcription len: 177.009085 (chars), i.e. ~3 chars per BPE label
    """

    def __init__(
        self,
        path,
        prefix,
        audio,
        orth_post_process=None,
        targets=None,
        chars=None,
        bpe=None,
        use_zip=False,
        use_ogg=False,
        use_cache_manager=False,
        fixed_random_subset=None,
        epoch_wise_filter=None,
        name=None,
        **kwargs,
    ):
        """
        :param str path: dir, should contain "train-*/*/*/{*.flac,*.trans.txt}", or "train-*.zip"
        :param str prefix: "train", "dev", "test", "dev-clean", "dev-other", ...
        :param str|list[str]|None orth_post_process: :func:`get_post_processor_function`, applied on orth
        :param str|dict[str]|None targets: "bpe" or "chars" or None or dict for :func:`Vocabulary.create_vocab`
        :param dict[str]|None audio: options for :class:`ExtractAudioFeatures`
        :param dict[str]|None bpe: options for :class:`BytePairEncoding`
        :param dict[str]|None chars: options for :class:`CharacterTargets`
        :param bool use_zip: whether to use the ZIP files instead (better for NFS)
        :param bool use_ogg: add .ogg postfix to all files
        :param bool use_cache_manager: uses :func:`Util.cf`
        :param float|int|None fixed_random_subset:
            Value in [0,1] to specify the fraction, or integer >=1 which specifies number of seqs.
            If given, will use this random subset. This will be applied initially at loading time,
            i.e. not dependent on the epoch. It will use an internally hardcoded fixed random seed,
            i.e. it's deterministic.
        :param dict|None epoch_wise_filter: see init_seq_order
        """
        if not name:
            name = "prefix:" + prefix
        super(LibriSpeechCorpus, self).__init__(name=name, **kwargs)
        import os
        from glob import glob
        import zipfile
        import returnn.util.basic

        self.path = path
        self.prefix = prefix
        self.use_zip = use_zip
        self.use_ogg = use_ogg
        self._zip_files = None
        if use_zip:
            zip_fn_pattern = "%s/%s*.zip" % (self.path, self.prefix)
            zip_fns = sorted(glob(zip_fn_pattern))
            assert zip_fns, "no files found: %r" % zip_fn_pattern
            if use_cache_manager:
                zip_fns = [returnn.util.basic.cf(fn) for fn in zip_fns]
            self._zip_files = {
                os.path.splitext(os.path.basename(fn))[0]: zipfile.ZipFile(fn) for fn in zip_fns
            }  # e.g. "train-clean-100" -> ZipFile
        assert prefix.split("-")[0] in ["train", "dev", "test"]
        assert os.path.exists(path + "/train-clean-100" + (".zip" if use_zip else ""))
        self.orth_post_process = None
        if orth_post_process:
            from .lm import get_post_processor_function

            self.orth_post_process = get_post_processor_function(orth_post_process)
        if isinstance(targets, dict):
            assert bpe is None and chars is None
            self.targets = Vocabulary.create_vocab(**targets)
        elif targets == "bpe" or (targets is None and bpe is not None):
            assert bpe is not None and chars is None
            self.targets = BytePairEncoding(**bpe)
        elif targets == "chars" or (targets is None and chars is not None):
            assert bpe is None and chars is not None
            self.targets = CharacterTargets(**chars)
        elif targets is None:
            assert bpe is None and chars is None
            self.targets = None  # type: typing.Optional[Vocabulary]
        else:
            raise Exception("invalid targets %r. provide bpe or chars" % targets)
        if self.targets:
            self.labels["classes"] = self.targets.labels
        self._audio_random = numpy.random.RandomState(1)
        self.feature_extractor = (
            ExtractAudioFeatures(random_state=self._audio_random, **audio) if audio is not None else None
        )
        self.num_inputs = self.feature_extractor.get_feature_dimension() if self.feature_extractor else 0
        self.num_outputs = {"raw": {"dtype": "string", "shape": ()}}
        if self.targets:
            self.num_outputs["classes"] = [self.targets.num_labels, 1]
        if self.feature_extractor:
            self.num_outputs["data"] = [self.num_inputs, 2]
        self.transs = self._collect_trans()
        self._reference_seq_order = sorted(self.transs.keys())
        if fixed_random_subset:
            if 0 < fixed_random_subset < 1:
                fixed_random_subset = int(len(self._reference_seq_order) * fixed_random_subset)
            assert isinstance(fixed_random_subset, int) and fixed_random_subset > 0
            rnd = numpy.random.RandomState(42)
            seqs = self._reference_seq_order
            rnd.shuffle(seqs)
            seqs = seqs[:fixed_random_subset]
            self._reference_seq_order = seqs
            self.transs = {s: self.transs[s] for s in seqs}
        self.epoch_wise_filter = epoch_wise_filter
        self._seq_order = None  # type: typing.Optional[typing.Sequence[int]]
        self.init_seq_order()

    def _collect_trans(self):
        from glob import glob
        import os
        import zipfile

        transs = (
            {}
        )  # type: typing.Dict[typing.Tuple[str,int,int,int],str]  # (subdir, speaker-id, chapter-id, seq-id) -> transcription  # nopep8
        if self.use_zip:
            for name, zip_file in self._zip_files.items():
                assert isinstance(zip_file, zipfile.ZipFile)
                assert zip_file.filelist
                assert zip_file.filelist[0].filename.startswith("LibriSpeech/")
                for info in zip_file.filelist:
                    assert isinstance(info, zipfile.ZipInfo)
                    path = info.filename.split("/")
                    assert path[0] == "LibriSpeech", "does not expect %r (%r)" % (info, info.filename)
                    if path[1].startswith(self.prefix):
                        subdir = path[1]  # e.g. "train-clean-100"
                        assert subdir == name
                        if path[-1].endswith(".trans.txt"):
                            for line in zip_file.read(info).decode("utf8").splitlines():
                                seq_name, txt = line.split(" ", 1)
                                speaker_id, chapter_id, seq_id = map(int, seq_name.split("-"))
                                if self.orth_post_process:
                                    txt = self.orth_post_process(txt)
                                transs[(subdir, speaker_id, chapter_id, seq_id)] = txt
        else:  # not zipped, directly read extracted files
            for subdir in glob("%s/%s*" % (self.path, self.prefix)):
                if not os.path.isdir(subdir):
                    continue
                subdir = os.path.basename(subdir)  # e.g. "train-clean-100"
                for fn in glob("%s/%s/*/*/*.trans.txt" % (self.path, subdir)):
                    for line in open(fn).read().splitlines():
                        seq_name, txt = line.split(" ", 1)
                        speaker_id, chapter_id, seq_id = map(int, seq_name.split("-"))
                        if self.orth_post_process:
                            txt = self.orth_post_process(txt)
                        transs[(subdir, speaker_id, chapter_id, seq_id)] = txt
            assert transs, "did not found anything %s/%s*" % (self.path, self.prefix)
        assert transs
        return transs

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        If random_shuffle_epoch1, for epoch 1 with "random" ordering, we leave the given order as is.
        Otherwise, this is mostly the default behavior.

        :param int|None epoch:
        :param list[str]|None seq_list: List of sequence tags, to set a predefined order.
        :param list[int]|None seq_order: List of corpus sequence indices, to set a predefined order.
        :rtype: bool
        :returns whether the order changed (True is always safe to return)
        """
        import returnn.util.basic

        super(LibriSpeechCorpus, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        if not epoch:
            epoch = 1
        random_seed = self._get_random_seed_for_epoch(epoch=epoch)
        self._audio_random.seed(random_seed)
        if self.targets:
            self.targets.set_random_seed(random_seed)

        def get_seq_len(i):
            """
            :param int i:
            :rtype: int
            """
            return len(self.transs[self._reference_seq_order[i]])

        if seq_order is not None:
            self._seq_order = seq_order
        elif seq_list is not None:
            seqs = [i for i in range(len(self._reference_seq_order)) if self._get_tag(i) in seq_list]
            seqs = {self._get_tag(i): i for i in seqs}
            for seq_tag in seq_list:
                assert seq_tag in seqs, "did not found all requested seqs. we have eg: %s" % (self._get_tag(0),)
            self._seq_order = [seqs[seq_tag] for seq_tag in seq_list]
        else:
            num_seqs = len(self._reference_seq_order)
            self._seq_order = self.get_seq_order_for_epoch(epoch=epoch, num_seqs=num_seqs, get_seq_len=get_seq_len)
        self._num_seqs = len(self._seq_order)
        if self.epoch_wise_filter:
            # Note: A more generic variant of this code is :class:`MetaDataset.EpochWiseFilter`.
            from .meta import EpochWiseFilter

            old_num_seqs = self._num_seqs
            any_filter = False
            for (ep_start, ep_end), value in sorted(self.epoch_wise_filter.items()):
                if ep_start is None:
                    ep_start = 1
                if ep_end is None or ep_end == -1:
                    ep_end = sys.maxsize
                assert isinstance(ep_start, int) and isinstance(ep_end, int) and 1 <= ep_start <= ep_end
                assert isinstance(value, dict)
                if ep_start <= epoch <= ep_end:
                    any_filter = True
                    opts = CollectionReadCheckCovered(value.copy())
                    if opts.get("subdirs") is not None:
                        subdirs = opts.get("subdirs", None)
                        assert isinstance(subdirs, list)
                        self._seq_order = [
                            idx for idx in self._seq_order if self._reference_seq_order[idx][0] in subdirs
                        ]
                        assert self._seq_order, "subdir filter %r invalid?" % (subdirs,)
                    if opts.get("use_new_filter"):
                        if "subdirs" in opts.collection:
                            opts.collection.pop("subdirs")
                        self._seq_order = EpochWiseFilter.filter_epoch(
                            opts=opts,
                            debug_msg_prefix="%s, epoch %i. " % (self, epoch),
                            get_seq_len=get_seq_len,
                            seq_order=self._seq_order,
                        )
                    else:
                        if opts.get("max_mean_len"):
                            max_mean_len = opts.get("max_mean_len")
                            seqs = numpy.array(
                                sorted(
                                    [(len(self.transs[self._reference_seq_order[idx]]), idx) for idx in self._seq_order]
                                )
                            )
                            # Note: This is somewhat incorrect. But keep the behavior,
                            # such that old setups are reproducible.
                            # You can use the option `use_new_filter` to get a better behavior.
                            num = returnn.util.basic.binary_search_any(
                                cmp=lambda num_: numpy.mean(seqs[:num_, 0]) > max_mean_len, low=1, high=len(seqs) + 1
                            )
                            assert num is not None
                            self._seq_order = list(seqs[:num, 1])
                            print(
                                (
                                    "%s, epoch %i."
                                    " Old mean seq len (transcription) is %f, new is %f, requested max is %f."
                                    " Old num seqs is %i, new num seqs is %i."
                                )
                                % (
                                    self,
                                    epoch,
                                    float(numpy.mean(seqs[:, 0])),
                                    float(numpy.mean(seqs[:num, 0])),
                                    max_mean_len,
                                    len(seqs),
                                    num,
                                ),
                                file=log.v4,
                            )
                    opts.assert_all_read()
                    self._num_seqs = len(self._seq_order)
            if any_filter:
                print(
                    "%s, epoch %i. Old num seqs %i, new num seqs %i." % (self, epoch, old_num_seqs, self._num_seqs),
                    file=log.v4,
                )
            else:
                print("%s, epoch %i. No filter for this epoch." % (self, epoch), file=log.v4)
        return True

    def get_current_seq_order(self):
        """
        :rtype: typing.Sequence[int]
        """
        assert self._seq_order is not None
        return self._seq_order

    def _get_ref_seq_idx(self, seq_idx):
        """
        :param int seq_idx:
        :return: idx in self._reference_seq_order
        :rtype: int
        """
        return self._seq_order[seq_idx]

    def have_corpus_seq_idx(self):
        """
        :rtype: bool
        """
        return True

    def get_corpus_seq_idx(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: int
        """
        return self._get_ref_seq_idx(seq_idx)

    def _get_tag(self, ref_seq_idx):
        """
        :param int ref_seq_idx:
        :rtype: str
        """
        subdir, speaker_id, chapter_id, seq_id = self._reference_seq_order[ref_seq_idx]
        return "%(sd)s-%(sp)i-%(ch)i-%(i)04i" % {"sd": subdir, "sp": speaker_id, "ch": chapter_id, "i": seq_id}

    def get_tag(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: str
        """
        return self._get_tag(self._get_ref_seq_idx(seq_idx))

    def get_all_tags(self):
        """
        :rtype: list[str]
        """
        return [self._get_tag(i) for i in range(len(self._reference_seq_order))]

    def get_total_num_seqs(self):
        """
        :rtype: int
        """
        return len(self._reference_seq_order)

    def _get_transcription(self, seq_idx):
        """
        :param int seq_idx:
        :return: (bpe, txt)
        :rtype: (list[int]|None, str)
        """
        seq_key = self._reference_seq_order[self._get_ref_seq_idx(seq_idx)]
        targets_txt = self.transs[seq_key]
        return self.targets.get_seq(targets_txt) if self.targets else None, targets_txt

    def _open_audio_file(self, seq_idx):
        """
        :param int seq_idx:
        :return: io.FileIO
        """
        import io
        import os
        import zipfile

        subdir, speaker_id, chapter_id, seq_id = self._reference_seq_order[self._get_ref_seq_idx(seq_idx)]
        audio_fn = "%(sd)s/%(sp)i/%(ch)i/%(sp)i-%(ch)i-%(i)04i.flac" % {
            "sd": subdir,
            "sp": speaker_id,
            "ch": chapter_id,
            "i": seq_id,
        }
        if self.use_ogg:
            audio_fn += ".ogg"
        if self.use_zip:
            audio_fn = "LibriSpeech/%s" % (audio_fn,)
            zip_file = self._zip_files[subdir]
            assert isinstance(zip_file, zipfile.ZipFile)
            raw_bytes = zip_file.read(audio_fn)
            return io.BytesIO(raw_bytes)
        else:
            audio_fn = "%s/%s" % (self.path, audio_fn)
            assert os.path.exists(audio_fn)
            return open(audio_fn, "rb")

    def _collect_single_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq
        """
        seq_tag = self.get_tag(seq_idx)
        if self.feature_extractor:
            with self._open_audio_file(seq_idx) as audio_file:
                features = self.feature_extractor.get_audio_features_from_raw_bytes(audio_file, seq_name=seq_tag)
        else:
            features = numpy.zeros(())  # currently the API requires some dummy values...
        bpe, txt = self._get_transcription(seq_idx)
        targets = {"raw": numpy.array(txt, dtype="object")}
        if bpe is not None:
            targets["classes"] = numpy.array(bpe, dtype="int32")
        return DatasetSeq(features=features, targets=targets, seq_idx=seq_idx, seq_tag=seq_tag)


class Enwik8Corpus(CachedDataset2):
    """
    enwik8
    """

    # Use a single HDF file, and cache it across all instances.
    _hdf_file = None

    def __init__(self, path, subset, seq_len, batch_num_seqs=None, subsubset=None, **kwargs):
        """
        :param str path:
        :param str subset: "training", "validation", "test"
        :param int seq_len:
        :param int|None batch_num_seqs: if given, will not shuffle the data but have it in such order,
          that with a given batch num_seqs setting, you could reuse the hidden state in an RNN
        :param int|(int,int)|None subsubset: end, (start,end), or full
        """
        assert subset in ["training", "validation", "test"]
        import os

        super(Enwik8Corpus, self).__init__(**kwargs)
        self.path = path
        assert os.path.isdir(path)
        self._prepare()
        self._unique = self._hdf_file.attrs["unique"]  # array label-idx -> byte idx (uint8, 0-255)
        labels = [bytes([b]) for b in self._unique]
        self.labels = {"data": labels, "classes": labels}
        self.num_inputs = len(labels)
        self.num_outputs = {"data": [self.num_inputs, 1], "classes": [self.num_inputs, 1]}
        self._data = self._hdf_file["split/%s/default" % subset]  # raw data, uint8 array
        if subsubset:
            if isinstance(subsubset, int):
                self._data = self._data[:subsubset]
            else:
                self._data = self._data[subsubset[0] : subsubset[1]]
        assert len(self._data) > 1
        self._seq_len = seq_len
        self._batch_num_seqs = batch_num_seqs
        self._random = numpy.random.RandomState(1)  # seed will be set in init_seq_order
        self._seq_starts = numpy.arange(0, len(self._data) - 1, seq_len)
        self._seq_order = None  # type: typing.Optional[typing.Sequence[int]]

    def get_data_dtype(self, key):
        """
        :param str key:
        :rtype: str
        """
        return "uint8"

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :param int epoch:
        :param list[str]|None seq_list:
        :param list[int]|None seq_order:
        :rtype: bool
        """
        super(Enwik8Corpus, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        if not epoch:
            epoch = 1
        epoch_part = None
        if self.partition_epoch:
            epoch_part = (epoch - 1) % self.partition_epoch
            epoch = ((epoch - 1) // self.partition_epoch) + 1
        self._random.seed(self._get_random_seed_for_epoch(epoch=epoch))
        self._num_seqs = len(self._seq_starts)
        self._num_timesteps = len(self._data) - 1
        if self._batch_num_seqs is None:
            self._seq_order = self.get_seq_order_for_epoch(
                epoch=epoch or 1, num_seqs=self._num_seqs, get_seq_len=lambda _: self._seq_len
            )
        else:
            if self._num_seqs % self._batch_num_seqs > 0:
                self._num_seqs -= self._num_seqs % self._batch_num_seqs
                self._num_timesteps = None
                assert self._num_seqs > 0
            assert self._num_seqs % self._batch_num_seqs == 0
            seq_index = numpy.array(list(range(self._num_seqs)))
            seq_index = seq_index.reshape((self._batch_num_seqs, self._num_seqs // self._batch_num_seqs))
            seq_index = seq_index.transpose()
            seq_index = seq_index.flatten()
            self._seq_order = seq_index
            if self.partition_epoch:
                assert self._num_seqs >= self.partition_epoch
                partition_epoch_num_seqs = [self._num_seqs // self.partition_epoch] * self.partition_epoch
                i = 0
                while sum(partition_epoch_num_seqs) < self._num_seqs:
                    partition_epoch_num_seqs[i] += 1
                    i += 1
                    assert i < self.partition_epoch
                assert sum(partition_epoch_num_seqs) == self._num_seqs
                self._num_seqs = partition_epoch_num_seqs[epoch_part]
                i = 0
                for n in partition_epoch_num_seqs[:epoch_part]:
                    i += n
                self._seq_order = seq_index[i : i + self._num_seqs]
        self._num_seqs = len(self._seq_order)
        return True

    def _collect_single_seq(self, seq_idx):
        idx = self._seq_order[seq_idx]
        src_seq_start = self._seq_starts[idx]
        tgt_seq_start = src_seq_start + 1
        tgt_seq_end = min(tgt_seq_start + self._seq_len, len(self._data))
        src_seq_end = tgt_seq_end - 1
        assert tgt_seq_end - tgt_seq_start == src_seq_end - src_seq_start > 0
        data = numpy.array(self._data[src_seq_start:tgt_seq_end], dtype="uint8")
        return DatasetSeq(
            seq_idx=seq_idx,
            features=data[:-1],
            targets=data[1:],
            seq_tag="offset_%i_%i" % (src_seq_start, src_seq_end - src_seq_start),
        )

    @property
    def _hdf_filename(self):
        return self.path + "/enwik8.hdf5"

    @property
    def _zip_filename(self):
        return self.path + "/enwik8.zip"

    def _prepare(self):
        """
        Reference:
        https://github.com/julian121266/RecurrentHighwayNetworks/blob/master/data/create_enwik8.py
        """
        if self._hdf_file:
            return
        import os
        import h5py

        if not os.path.exists(self._hdf_filename):
            self._create_hdf()
        Enwik8Corpus._hdf_file = h5py.File(self._hdf_filename, "r")

    def _create_hdf(self):
        import os
        import h5py
        import zipfile

        if not os.path.exists(self._zip_filename):
            self._download_zip()

        print("%s: create %s" % (self, self._hdf_filename), file=log.v2)
        num_test_chars = 5000000

        raw_data = zipfile.ZipFile(self._zip_filename).read("enwik8").decode("utf8")
        raw_data = numpy.fromstring(raw_data, dtype=numpy.uint8)
        unique, data = numpy.unique(raw_data, return_inverse=True)

        train_data = data[: -2 * num_test_chars]
        valid_data = data[-2 * num_test_chars : -num_test_chars]
        test_data = data[-num_test_chars:]

        f = h5py.File(self._hdf_filename, "w")
        f.attrs["unique"] = unique

        variant = f.create_group("split")
        group = variant.create_group("training")
        group.create_dataset(name="default", data=train_data, compression="gzip")

        group = variant.create_group("validation")
        group.create_dataset(name="default", data=valid_data, compression="gzip")

        group = variant.create_group("test")
        group.create_dataset(name="default", data=test_data, compression="gzip")

        f.close()

    def _download_zip(self):
        url = "https://mattmahoney.net/dc/enwik8.zip"
        print("%s: download %s" % (self, url), file=log.v2)
        # noinspection PyPackageRequirements
        from six.moves.urllib.request import urlretrieve

        urlretrieve(url, self._zip_filename)


def demo():
    """
    Some demo for some of the :class:`GeneratingDataset`.
    """
    from returnn.util import better_exchook

    better_exchook.install()
    log.initialize(verbosity=[5])
    import sys

    dsclazzeval = sys.argv[1]
    dataset = eval(dsclazzeval)
    assert isinstance(dataset, Dataset)
    assert isinstance(dataset, GeneratingDataset), "use tools/dump-dataset.py for a generic demo instead"
    # noinspection PyProtectedMember
    assert dataset._input_classes and dataset._output_classes
    assert dataset.num_outputs["data"][1] == 2  # expect 1-hot
    assert dataset.num_outputs["classes"][1] == 1  # expect sparse
    for i in range(10):
        print("Seq idx %i:" % i)
        s = dataset.generate_seq(i)
        assert isinstance(s, DatasetSeq)
        features = s.features["data"]
        output_seq = s.features["classes"]
        assert features.ndim == 2
        assert output_seq.ndim == 1
        input_seq = numpy.argmax(features, axis=1)
        # noinspection PyProtectedMember
        input_seq_str = "".join([dataset._input_classes[i] for i in input_seq])
        # noinspection PyProtectedMember
        output_seq_str = "".join([dataset._output_classes[i] for i in output_seq])
        print(" %r" % input_seq_str)
        print(" %r" % output_seq_str)
        assert features.shape[1] == dataset.num_outputs["data"][0]
        assert features.shape[0] == output_seq.shape[0]


if __name__ == "__main__":
    demo()



"""
Provides the learning rate scheduling logic.
The base class is :class:`LearningRateControl`.
"""

from __future__ import annotations

import os
import typing
from returnn.util.basic import better_repr, simple_obj_repr, ObjAsDict, unicode
from returnn.log import log
import numpy


class LearningRateControl(object):
    """
    Base class for learning rate control / scheduling.
    """

    need_error_info = True

    class EpochData:
        """
        Encapsulates all relevant information for one epoch,
        needed to perform learning rate scheduling,
        such as the individual scores (cv or train; cross-entropy or frame-error or whatever).
        """

        # Need to keep the non-PEP8 name for compatibility, because we store the repr of the object.
        # noinspection PyPep8Naming
        def __init__(self, learningRate, error=None):
            """
            :type learningRate: float
            :type error: dict[str,float] | None
            """
            self.learning_rate = learningRate
            if isinstance(error, float):  # Old format.
                error = {"old_format_score": error}
            if error is None:
                error = {}
            self.error = error

        def __repr__(self):
            # This is being used for serialization, and we want some forward/backward compatibility,
            # so we should try to keep this consistent.
            return "EpochData(learningRate=%s, error=%s)" % (better_repr(self.learning_rate), better_repr(self.error))

    @classmethod
    def load_initial_kwargs_from_config(cls, config):
        """
        :type config: returnn.config.Config
        :rtype: dict[str]
        """
        return {
            "default_learning_rate": config.float("learning_rate", 1.0),
            "min_learning_rate": config.float("min_learning_rate", 0.0),
            "default_learning_rates": config.typed_value("learning_rates") or config.float_list("learning_rates"),
            "error_measure_key": (
                config.typed_value("learning_rate_control_error_measure")
                or config.value("learning_rate_control_error_measure", None)
            ),
            "relative_error_also_relative_to_learning_rate": (
                config.bool("learning_rate_control_relative_error_relative_lr", False)
            ),
            "min_num_epochs_per_new_learning_rate": config.int("learning_rate_control_min_num_epochs_per_new_lr", 0),
            "relative_error_div_by_old": config.bool("newbob_relative_error_div_by_old", False),
            "learning_rate_decay": config.typed_value(
                "learning_rate_decay", config.opt_typed_value("newbob_learning_rate_decay", 0.5)
            ),
            "learning_rate_growth": config.typed_value(
                "learning_rate_growth", config.opt_typed_value("newbob_learning_rate_growth", 1.0)
            ),
            "filename": config.value("learning_rate_file", None),
        }

    @classmethod
    def load_initial_from_config(cls, config):
        """
        :type config: returnn.config.Config
        :rtype: LearningRateControl
        """
        kwargs = cls.load_initial_kwargs_from_config(config)
        return cls(**kwargs)

    def __init__(
        self,
        default_learning_rate,
        min_learning_rate=0.0,
        default_learning_rates=None,
        error_measure_key=None,
        relative_error_also_relative_to_learning_rate=False,
        min_num_epochs_per_new_learning_rate=0,
        relative_error_div_by_old=False,
        learning_rate_decay=1.0,
        learning_rate_growth=1.0,
        filename=None,
    ):
        """
        :param float default_learning_rate: default learning rate. usually for epoch 1
        :param list[float] | dict[int,float] default_learning_rates: learning rates
        :param str|list[str]|None error_measure_key: for get_epoch_error_value() the key for EpochData.error
            which is a dict
        :param int min_num_epochs_per_new_learning_rate: if the lr was recently updated, use it for at least N epochs
        :param bool relative_error_div_by_old: if True, compute relative error as (new - old) / old.
        :param float|(float)->float learning_rate_decay:
        :param float|(float)->float learning_rate_growth:
        :param str filename: load from and save to file
        """
        self.epoch_data = {}  # type: typing.Dict[int,LearningRateControl.EpochData]
        self.filename = filename
        if filename:
            if os.path.exists(filename):
                print("Learning-rate-control: loading file %s" % filename, file=log.v4)
                # Load now, such that default_learning_rates is correctly handled.
                self.load()
            else:
                print("Learning-rate-control: file %s does not exist yet" % filename, file=log.v4)
        else:
            print(
                "Learning-rate-control: no file specified, not saving history (no proper restart possible)", file=log.v4
            )
        self.default_learning_rate = default_learning_rate
        self.min_learning_rate = min_learning_rate
        if default_learning_rates:
            if isinstance(default_learning_rates, list):
                default_learning_rates = {i + 1: v for (i, v) in enumerate(default_learning_rates)}
            if isinstance(default_learning_rates, (str, unicode)):
                default_learning_rates = eval(default_learning_rates)
            assert isinstance(default_learning_rates, dict)
            for epoch, v in default_learning_rates.items():
                self.set_default_learning_rate_for_epoch(epoch, v)
        self.default_learning_rates = default_learning_rates
        self.error_measure_key = error_measure_key
        self.relative_error_also_relative_to_learning_rate = relative_error_also_relative_to_learning_rate
        self.min_num_epochs_per_new_learning_rate = min_num_epochs_per_new_learning_rate
        self.relative_error_div_by_old = relative_error_div_by_old
        self.learning_rate_decay = learning_rate_decay
        self.learning_rate_growth = learning_rate_growth

    __repr__ = simple_obj_repr

    def __str__(self):
        epochs = sorted(self.epoch_data.keys())
        if len(epochs) > 6:
            epoch_str = ", ".join(
                ["%i: %s" % (epoch, self.epoch_data[epoch]) for epoch in epochs[:3]]
                + ["..."]
                + ["%i: %s" % (epoch, self.epoch_data[epoch]) for epoch in epochs[-3:]]
            )
        else:
            epoch_str = ", ".join(["%i: %s" % (epoch, self.epoch_data[epoch]) for epoch in epochs])
        return "%r, epoch data: %s, error key: %s" % (self, epoch_str, self.get_error_key(epoch=1))

    @staticmethod
    def _calc_learning_rate_update(learning_rate, update):
        """
        :param float learning_rate:
        :param None|float|(float)->float update: factor, or generic func
        :return: lr with update applied (e.g. decay factor)
        :rtype: float
        """
        if update is None:
            return learning_rate
        if isinstance(update, float):
            return learning_rate * update
        assert callable(update)
        learning_rate = update(learning_rate)
        assert isinstance(learning_rate, float)
        return learning_rate

    def _calc_learning_rate_decay(self, learning_rate):
        """
        :param float learning_rate:
        :return: lr with decay applied
        :rtype: float
        """
        return self._calc_learning_rate_update(learning_rate, update=self.learning_rate_decay)

    def _calc_learning_rate_growth(self, learning_rate):
        """
        :param float learning_rate:
        :return: lr with growth applied
        :rtype: float
        """
        return self._calc_learning_rate_update(learning_rate, update=self.learning_rate_growth)

    def calc_learning_rate_decay_or_grow(self, learning_rate, decay, grow=None):
        """
        :param float learning_rate:
        :param bool decay:
        :param bool|None grow: default is not decay
        :return: lr with decay or growth applied
        :rtype: float
        """
        assert isinstance(decay, bool)
        if grow is None:
            grow = not decay
        assert isinstance(grow, bool)
        assert not (grow and decay)  # not sure if this makes sense...
        if decay:
            learning_rate = self._calc_learning_rate_decay(learning_rate)
            if learning_rate < self.min_learning_rate:
                learning_rate = self.min_learning_rate
        if grow:
            learning_rate = self._calc_learning_rate_growth(learning_rate)
        return learning_rate

    def calc_learning_rate_for_epoch(self, epoch):
        """
        :type epoch: int
        :returns learning rate
        :rtype: float
        """
        raise NotImplementedError

    def calc_new_learning_rate_for_epoch(self, epoch):
        """
        :param int epoch:
        :return: new learning rate for this epoch
        :rtype: float
        """
        if self.min_num_epochs_per_new_learning_rate > 1:
            last_lrs = [
                self.epoch_data[e].learning_rate
                for e in self._last_epochs_for_epoch(epoch, num_epochs=self.min_num_epochs_per_new_learning_rate)
            ]
            if len(set(last_lrs)) >= 2 or 0 < len(last_lrs) < self.min_num_epochs_per_new_learning_rate:
                return last_lrs[-1]
        learning_rate = self.calc_learning_rate_for_epoch(epoch)
        return learning_rate

    def _last_epochs_for_epoch(self, epoch, num_epochs):
        """
        :param int epoch:
        :param int num_epochs:
        :return: last N epochs where we have some epoch data
        :rtype: list[int]
        """
        last_epochs = sorted([e for e in self.epoch_data.keys() if e < epoch])
        if not last_epochs:
            return []
        last_epochs = last_epochs[-num_epochs:]
        return last_epochs

    def get_learning_rate_for_epoch(self, epoch):
        """
        :type epoch: int
        :rtype: float
        """
        assert epoch >= 1
        if epoch in self.epoch_data:
            return self.epoch_data[epoch].learning_rate
        learning_rate = self.calc_new_learning_rate_for_epoch(epoch)
        self.set_default_learning_rate_for_epoch(epoch, learning_rate)
        return learning_rate

    def set_default_learning_rate_for_epoch(self, epoch, learning_rate):
        """
        :type epoch: int
        :type learning_rate: float
        """
        if epoch in self.epoch_data:
            if not self.epoch_data[epoch].learning_rate:
                self.epoch_data[epoch].learning_rate = learning_rate
        else:
            self.epoch_data[epoch] = self.EpochData(learning_rate)

    def get_last_epoch(self, epoch):
        """
        :param int epoch:
        :return: last epoch before ``epoch`` where we have some epoch data
        :rtype: int
        """
        epochs = sorted([e for e in self.epoch_data.keys() if e < epoch])
        if not epochs:
            return None
        return epochs[-1]

    def get_most_recent_learning_rate(self, epoch, exclude_current=True):
        """
        :param int epoch:
        :param bool exclude_current:
        :return: most learning rate before or including ``epoch``
        :rtype: float
        """
        for e, data in reversed(sorted(self.epoch_data.items())):
            assert isinstance(data, LearningRateControl.EpochData)
            if e > epoch:
                continue
            if exclude_current and e == epoch:
                continue
            if data.learning_rate is None:
                continue
            return data.learning_rate
        return self.default_learning_rate

    def calc_relative_error(self, old_epoch, new_epoch):
        """
        :param int old_epoch:
        :param int new_epoch:
        :return: relative error between old epoch and new epoch
        :rtype: float
        """
        old_key, old_error = self.get_epoch_error_key_value(old_epoch)
        new_key, new_error = self.get_epoch_error_key_value(new_epoch)
        if old_error is None or new_error is None:
            return None
        if old_key != new_key:
            return None
        if self.relative_error_div_by_old:
            relative_error = (new_error - old_error) / abs(old_error)
        else:
            relative_error = (new_error - old_error) / abs(new_error)
        if self.relative_error_also_relative_to_learning_rate:
            learning_rate = self.get_most_recent_learning_rate(new_epoch, exclude_current=False)
            if learning_rate > 0:
                # If the learning rate is lower than the initial learning rate,
                # the relative error is also expected to be lower, so correct for that here.
                relative_error /= learning_rate / self.default_learning_rate
        return relative_error

    def set_epoch_error(self, epoch, error):
        """
        :type epoch: int
        :type error: dict[str,float|dict[str,float]]
        """
        if epoch not in self.epoch_data:
            print("Learning rate not set for epoch %i. Assuming default." % epoch, file=log.v4)
            self.get_learning_rate_for_epoch(epoch)  # This will set it.
        assert isinstance(error, dict)
        error = error.copy()
        for k, v in list(error.items()):
            if isinstance(v, dict):  # like error = {"dev_score": {"cost:output1": .., "cost:output2": ...}, ...}
                del error[k]
                if len(v) == 1:
                    error[k] = list(v.values())[0]
                    continue
                for k1, v1 in v.items():
                    if ":" in k1:
                        k1 = k1[k1.index(":") + 1 :]
                    error[k + "_" + k1] = v1
        for v in error.values():
            assert isinstance(v, float)
        self.epoch_data[epoch].error.update(error)
        if epoch == 1:
            print("Learning-rate-control: error key %r from %r" % (self.get_error_key(epoch), error), file=log.v4)

    def get_error_key(self, epoch):
        """
        :param int epoch:
        :return: key which we should look in scores/errors, for this epoch
        :rtype: str
        """
        if epoch not in self.epoch_data:
            if isinstance(self.error_measure_key, list):
                return self.error_measure_key[0]
            assert isinstance(self.error_measure_key, (str, type(None)))
            return self.error_measure_key
        epoch_data = self.epoch_data[epoch]
        if not epoch_data.error:
            return None
        if len(epoch_data.error) == 1 and "old_format_score" in epoch_data.error:
            return "old_format_score"
        keys = []
        if isinstance(self.error_measure_key, list):
            for key in self.error_measure_key:
                keys += [key, key + "_output"]  # for multiple outputs, try default output
        elif isinstance(self.error_measure_key, str):
            keys += [self.error_measure_key, self.error_measure_key + "_output"]
        else:
            assert self.error_measure_key is None
        keys += ["dev_score", "dev_score_output"]
        for key in keys:
            if key in epoch_data.error:
                return key
        for key in sorted(epoch_data.error.keys()):
            if key == "dev_score_output/output" or key.startswith("dev_score_output/output_"):
                return key
        for key in sorted(epoch_data.error.keys()):
            if key.startswith("dev_score_output/"):
                return key
        for key in sorted(epoch_data.error.keys()):
            if key.startswith("dev_"):
                return key
        for key in ["train_score", "train_score_output"]:
            if key in epoch_data.error:
                return key
        return min(epoch_data.error.keys())

    def get_epoch_error_dict(self, epoch):
        """
        :param int epoch:
        :rtype: dict[str,float]
        """
        if epoch not in self.epoch_data:
            return {}
        return self.epoch_data[epoch].error

    def get_epoch_error_value(self, epoch):
        """
        :param int epoch:
        :return: error/score for the specific epoch, given the error-key, see :func:`get_error_key`
        :rtype: float
        """
        error = self.get_epoch_error_dict(epoch)
        if not error:
            return None
        key = self.get_error_key(epoch)
        assert key
        assert key in error, "%r not in %r. fix %r in config. set it to %r or so." % (
            key,
            error,
            "learning_rate_control_error_measure",
            "dev_error",
        )
        return error[key]

    def get_epoch_error_key_value(self, epoch):
        """
        :param int epoch:
        :return: key, error
        :rtype: (str, float)
        """
        error = self.get_epoch_error_dict(epoch)
        if not error:
            return None, None
        key = self.get_error_key(epoch)
        assert key
        assert key in error, "%r not in %r. fix %r in config. set it to %r or so." % (
            key,
            error,
            "learning_rate_control_error_measure",
            "dev_error",
        )
        return key, error[key]

    def get_last_best_epoch(
        self,
        last_epoch,
        first_epoch=1,
        only_last_epochs=None,
        filter_score=float("inf"),
        only_last_n=-1,
        min_score_dist=0.0,
    ):
        """
        :param int first_epoch: will check all epochs >= first_epoch
        :param int last_epoch: inclusive. will check all epochs <= last_epoch
        :param int|None only_last_epochs: if set, will only check the last N epochs, inclusive
        :param float filter_score: all epochs which values over this score are not considered
        :param int only_last_n: if set (>=1), *from the resulting list*, we consider only the last only_last_n
        :param float min_score_dist: filter out epochs where the diff to the most recent is not big enough
        :return: the last best epoch. to get the details then, you might want to use getEpochErrorDict.
        :rtype: int|None
        """
        if only_last_epochs:
            first_epoch = max(first_epoch, last_epoch - only_last_epochs + 1)
        if first_epoch > last_epoch:
            return None
        values = [(self.get_epoch_error_key_value(ep), ep) for ep in range(first_epoch, last_epoch + 1)]
        # Note that the order of the checks here is a bit arbitrary but I had some thoughts on it.
        # Changing the order will also slightly change the behavior, so be sure it make sense.
        values = [((key, v), ep) for ((key, v), ep) in values if v is not None]
        if not values:
            return None
        last_key, latest_score = values[-1][0]
        values = [(v, ep) for ((key, v), ep) in values if key == last_key]  # only same key
        values = [(v, ep) for (v, ep) in values if v <= filter_score]
        if not values:
            return None
        if only_last_n >= 1:
            values = values[-only_last_n:]
        values = [(v, ep) for (v, ep) in values if v + min_score_dist < latest_score]
        if not values:
            return None
        return min(values)[1]

    def save(self):
        """
        Save the current epoch data to file (self.filename).
        """
        if not self.filename:
            return
        # First write to a temp-file, to be sure that the write happens without errors.
        # Otherwise, it could happen that we delete the old existing file, then
        # some error happens (e.g. disk quota), and we loose the newbob data.
        # Loosing that data is very bad because it basically means that we have to redo all the training.
        tmp_filename = self.filename + ".new_tmp"
        f = open(tmp_filename, "w")
        f.write(better_repr(self.epoch_data))
        f.write("\n")
        f.close()
        os.rename(tmp_filename, self.filename)

    def load(self):
        """
        Loads the saved epoch data from file (self.filename).
        """
        s = open(self.filename).read()
        self.epoch_data = eval(s, {"nan": float("nan"), "inf": float("inf")}, ObjAsDict(self))


class ConstantLearningRate(LearningRateControl):
    """
    Just a constant learning rate.
    """

    need_error_info = False

    def calc_learning_rate_for_epoch(self, epoch):
        """
        Dummy constant learning rate. Returns initial learning rate.
        :type epoch: int
        :returns learning rate
        :rtype: float
        """
        while True:
            last_epoch = self.get_last_epoch(epoch)
            if last_epoch is None:
                return self.default_learning_rate
            learning_rate = self.epoch_data[last_epoch].learning_rate
            if learning_rate is None:
                epoch = last_epoch
                continue
            return learning_rate


class NewbobRelative(LearningRateControl):
    """
    If relative diff between old and new error is over some threshold, decay learning rate.
    """

    @classmethod
    def load_initial_kwargs_from_config(cls, config):
        """
        :type config: returnn.config.Config
        :rtype: dict[str]
        """
        kwargs = super(NewbobRelative, cls).load_initial_kwargs_from_config(config)
        kwargs.update(
            {
                "relative_error_threshold": config.float("newbob_relative_error_threshold", -0.01),
            }
        )
        return kwargs

    def __init__(self, relative_error_threshold, **kwargs):
        """
        :type relative_error_threshold: float
        """
        super(NewbobRelative, self).__init__(**kwargs)
        self.relative_error_threshold = relative_error_threshold

    def calc_learning_rate_for_epoch(self, epoch):
        """
        Newbob+ on train data.
        :type epoch: int
        :returns learning rate
        :rtype: float
        """
        last_epoch = self.get_last_epoch(epoch)
        if last_epoch is None:
            return self.default_learning_rate
        learning_rate = self.epoch_data[last_epoch].learning_rate
        if learning_rate is None:
            return self.default_learning_rate
        last2_epoch = self.get_last_epoch(last_epoch)
        if last2_epoch is None:
            return learning_rate
        relative_error = self.calc_relative_error(last2_epoch, last_epoch)
        if relative_error is None:
            return learning_rate
        learning_rate = self.calc_learning_rate_decay_or_grow(
            learning_rate, decay=relative_error > self.relative_error_threshold
        )
        return learning_rate


class NewbobAbs(LearningRateControl):
    """
    If absolute diff between old and new error is over some threshold, decay learning rate.
    """

    @classmethod
    def load_initial_kwargs_from_config(cls, config):
        """
        :type config: returnn.config.Config
        :rtype: dict[str]
        """
        kwargs = super(NewbobAbs, cls).load_initial_kwargs_from_config(config)
        kwargs.update(
            {
                "error_threshold": config.float("newbob_error_threshold", -0.01),
            }
        )
        return kwargs

    def __init__(self, error_threshold, **kwargs):
        """
        :type error_threshold: float
        """
        super(NewbobAbs, self).__init__(**kwargs)
        self.error_threshold = error_threshold

    def calc_learning_rate_for_epoch(self, epoch):
        """
        Newbob+ on train data.

        :type epoch: int
        :returns learning rate
        :rtype: float
        """
        last_epoch = self.get_last_epoch(epoch)
        if last_epoch is None:
            return self.default_learning_rate
        learning_rate = self.epoch_data[last_epoch].learning_rate
        if learning_rate is None:
            return self.default_learning_rate
        last2_epoch = self.get_last_epoch(last_epoch)
        if last2_epoch is None:
            return learning_rate
        old_key, old_error = self.get_epoch_error_key_value(last2_epoch)
        new_key, new_error = self.get_epoch_error_key_value(last_epoch)
        if old_error is None or new_error is None:
            return learning_rate
        if old_key != new_key:
            return learning_rate
        error_diff = new_error - old_error
        learning_rate = self.calc_learning_rate_decay_or_grow(learning_rate, decay=error_diff > self.error_threshold)
        return learning_rate


class NewbobMultiEpoch(LearningRateControl):
    """
    Like :class:`NewbobRelative`, but looks at the average relative error over multiple epochs.
    This is useful together with ``partition_epoch`` from :class:`Dataset`.
    """

    @classmethod
    def load_initial_kwargs_from_config(cls, config):
        """
        :type config: returnn.config.Config
        :rtype: dict[str]
        """
        kwargs = super(NewbobMultiEpoch, cls).load_initial_kwargs_from_config(config)
        kwargs.update(
            {
                "num_epochs": config.int("newbob_multi_num_epochs", 5),
                "update_interval": config.int("newbob_multi_update_interval", config.int("newbob_multi_num_epochs", 5)),
                "relative_error_threshold": config.float("newbob_relative_error_threshold", -0.01),
                "relative_error_grow_threshold": config.float(
                    "newbob_relative_error_grow_threshold", config.float("newbob_relative_error_threshold", -0.01)
                ),
            }
        )
        return kwargs

    def __init__(self, num_epochs, update_interval, relative_error_threshold, relative_error_grow_threshold, **kwargs):
        """
        :param int num_epochs:
        :param int update_interval:
        :param float relative_error_threshold:
        :param float relative_error_grow_threshold:
        """
        super(NewbobMultiEpoch, self).__init__(**kwargs)
        self.num_epochs = num_epochs
        assert self.num_epochs >= 1
        self.update_interval = update_interval
        assert self.update_interval >= 1
        self.relative_error_threshold = relative_error_threshold
        self.relative_error_grow_threshold = relative_error_grow_threshold

    def _calc_mean_relative_error(self, epochs):
        """
        :param list[int] epochs:
        :return: mean of relative errors
        :rtype: float|None
        """
        assert len(epochs) >= 2
        errors = [self.calc_relative_error(epochs[i], epochs[i + 1]) for i in range(len(epochs) - 1)]
        if any([e is None for e in errors]):
            return None
        return float(numpy.mean(errors))

    def _calc_recent_mean_relative_error(self, epoch):
        """
        :param int epoch:
        :return: recent mean of relative errors
        :rtype: float|None
        """
        # Take one more than numEpochs because we are looking at the diffs.
        last_epochs = self._last_epochs_for_epoch(epoch, num_epochs=self.num_epochs + 1)
        if not last_epochs:
            return None
        # We could also use the self.numEpochs limit here. But maybe this is better.
        if len(last_epochs) <= 1:
            return None
        return self._calc_mean_relative_error(last_epochs)

    def calc_learning_rate_for_epoch(self, epoch):
        """
        Newbob+ on train data.
        :type epoch: int
        :returns learning rate
        :rtype: float
        """
        learning_rate = self.get_most_recent_learning_rate(epoch)
        # We start counting epochs at 1.
        if self.update_interval > 1 and epoch % self.update_interval != 1:
            return learning_rate
        mean_relative_error = self._calc_recent_mean_relative_error(epoch)
        if mean_relative_error is None:
            return learning_rate
        learning_rate = self.calc_learning_rate_decay_or_grow(
            learning_rate,
            decay=mean_relative_error > self.relative_error_threshold,
            grow=mean_relative_error < self.relative_error_grow_threshold,
        )
        return learning_rate


def learning_rate_control_type(type_name):
    """
    :param str type_name:
    :rtype: type[LearningRateControl]|LearningRateControl
    """
    if type_name == "constant":
        return ConstantLearningRate
    elif type_name in ("newbob", "newbob_rel", "newbob_relative"):  # Old setups expect the relative version.
        return NewbobRelative
    elif type_name == "newbob_abs":
        return NewbobAbs
    elif type_name == "newbob_multi_epoch":
        return NewbobMultiEpoch
    else:
        assert False, "unknown learning-rate-control type %s" % type_name


def load_learning_rate_control_from_config(config):
    """
    :type config: returnn.config.Config
    :rtype: LearningRateControl
    """
    control_type = config.value("learning_rate_control", "constant")
    cls = learning_rate_control_type(control_type)
    return cls.load_initial_from_config(config)


def demo():
    """
    Demo run. Given some learning rate file (with scores / existing lrs), will calculate how lrs would have been set,
    given some config.
    """
    from returnn.util import better_exchook

    better_exchook.install()
    import returnn.__main__ as rnn
    import sys

    if len(sys.argv) <= 1:
        print("usage: python %s [config] [other options] [++check_learning_rates 1]" % __file__)
        print(
            (
                "example usage: "
                "python %s ++learning_rate_control newbob ++learning_rate_file newbob.data ++learning_rate 0.001"
            )
            % __file__
        )
    rnn.init_config(command_line_options=sys.argv[1:])
    # noinspection PyProtectedMember
    rnn.config._hack_value_reading_debug()
    rnn.config.update({"log": []})
    rnn.init_log()
    rnn.init_backend_engine()
    check_lr = rnn.config.bool("check_learning_rates", False)
    from returnn.pretrain import pretrain_from_config

    pretrain = pretrain_from_config(rnn.config)
    first_non_pretrain_epoch = 1
    pretrain_learning_rate = None
    if pretrain:
        first_non_pretrain_epoch = pretrain.get_train_num_epochs() + 1
    log.initialize(verbosity=[5])
    control = load_learning_rate_control_from_config(rnn.config)
    print("LearningRateControl: %r" % control)
    if not control.epoch_data:
        print("No epoch data so far.")
        return
    first_epoch = min(control.epoch_data.keys())
    if first_epoch != 1:
        print("Strange, first epoch from epoch data is %i." % first_epoch)
    print("Error key: %s from %r" % (control.get_error_key(epoch=first_epoch), control.epoch_data[first_epoch].error))
    if pretrain:
        pretrain_learning_rate = rnn.config.float("pretrain_learning_rate", control.default_learning_rate)
    max_epoch = max(control.epoch_data.keys())
    for epoch in range(1, max_epoch + 2):  # all epochs [1..max_epoch+1]
        old_learning_rate = None
        if epoch in control.epoch_data:
            old_learning_rate = control.epoch_data[epoch].learning_rate
        if epoch < first_non_pretrain_epoch:
            learning_rate = pretrain_learning_rate
            s = "Pretrain epoch %i, fixed learning rate: %s (was: %s)" % (epoch, learning_rate, old_learning_rate)
        elif 1 < first_non_pretrain_epoch == epoch:
            learning_rate = control.default_learning_rate
            s = "First epoch after pretrain, epoch %i, fixed learning rate: %s (was %s)" % (
                epoch,
                learning_rate,
                old_learning_rate,
            )
        else:
            learning_rate = control.calc_new_learning_rate_for_epoch(epoch)
            s = "Calculated learning rate for epoch %i: %s (was: %s)" % (epoch, learning_rate, old_learning_rate)
        if learning_rate < control.min_learning_rate:
            learning_rate = control.min_learning_rate
            s += ", clipped to %s" % learning_rate
        s += ", previous relative error: %s" % control.calc_relative_error(epoch - 2, epoch - 1)
        if hasattr(control, "_calc_recent_mean_relative_error"):
            # noinspection PyProtectedMember
            s += ", previous mean relative error: %s" % control._calc_recent_mean_relative_error(epoch)
        print(s)
        if check_lr and old_learning_rate is not None:
            if old_learning_rate != learning_rate:
                print("Learning rate is different in epoch %i!" % epoch)
                sys.exit(1)
        # Overwrite new learning rate so that the calculation for further learning rates stays consistent.
        if epoch in control.epoch_data:
            control.epoch_data[epoch].learning_rate = learning_rate
        else:
            control.epoch_data[epoch] = control.EpochData(learningRate=learning_rate)
    print("Finished, last stored epoch was %i." % max_epoch)


if __name__ == "__main__":
    demo()



"""
Common parameter initialization functions.

https://github.com/rwth-i6/returnn/wiki/Parameter-initialization
"""

from __future__ import annotations
from typing import Union, Sequence, Optional
from returnn.tensor import Tensor, Dim
import math
from .. import frontend as rf


class ParamInit:
    """API for param init"""

    def __call__(self, shape: Sequence[Dim], dtype: str) -> Union[Tensor, rf.RawTensorTypes]:
        raise NotImplementedError


ParamInitType = Union[Tensor, rf.RawTensorTypes, ParamInit]


class VarianceScaling(ParamInit):
    """
    Provides a generalized way for initializing weights.
    All the common initialization methods are special cases
    such as Xavier Glorot and Kaiming He.

    Code adopted from TensorFlow VarianceScaling.
    """

    scale = 1.0
    mode = "fan_in"  # fan_in, fan_out, fan_avg
    distribution = "truncated_normal"  # normal, untruncated_normal, truncated_normal, uniform
    dtype = "float32"

    def __init__(self, scale: float = None, mode: str = None, distribution: str = None, dtype: str = None):
        if scale is not None:
            self.scale = scale
        if mode is not None:
            self.mode = mode
        if distribution is not None:
            self.distribution = distribution
        if dtype is not None:
            self.dtype = dtype

        if self.scale <= 0.0:
            raise ValueError(f"Argument `scale` must be a positive float. Received: {self.scale}")
        if self.mode not in {"fan_in", "fan_out", "fan_avg"}:
            raise ValueError(
                f"Argument `mode` should be one of ('fan_in', 'fan_out', 'fan_avg'). Received: {self.mode}"
            )
        if self.distribution not in {"normal", "uniform", "truncated_normal", "untruncated_normal"}:
            raise ValueError(
                "Argument `distribution` should be one of "
                "('normal', 'uniform', 'truncated_normal', 'untruncated_normal'). "
                f"Received: {self.distribution}"
            )

    def __call__(self, shape: Sequence[Dim], dtype: Optional[str] = None) -> Tensor:
        if dtype is None:
            dtype = self.dtype
        scale = self.scale
        fan_in, fan_out = _compute_fans(shape)
        if self.mode == "fan_in":
            scale /= max(1.0, fan_in)
        elif self.mode == "fan_out":
            scale /= max(1.0, fan_out)
        else:
            scale /= max(1.0, (fan_in + fan_out) / 2.0)
        return self._random(shape=shape, dtype=dtype, scale=scale)

    def _random(self, shape: Sequence[Dim], scale: float, dtype=None) -> Tensor:
        if self.distribution in {"truncated_normal", "normal"}:
            # constant taken from scipy.stats.truncnorm.std(a=-2, b=2, loc=0., scale=1.)
            stddev = math.sqrt(scale) / 0.87962566103423978
            return rf.random(
                distribution=self.distribution, static=True, shape=shape, mean=0.0, stddev=stddev, dtype=dtype
            )
        elif self.distribution == "untruncated_normal":
            stddev = math.sqrt(scale)
            return rf.random(
                distribution=self.distribution, static=True, shape=shape, mean=0.0, stddev=stddev, dtype=dtype
            )
        elif self.distribution == "uniform":
            limit = math.sqrt(3.0 * scale)
            return rf.random(
                distribution=self.distribution, static=True, shape=shape, minval=-limit, maxval=limit, dtype=dtype
            )
        else:
            raise ValueError(f"invalid distribution {self.distribution!r}")


class Glorot(VarianceScaling):
    """
    Xavier Glorot (http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf).
    scale 1, fan_avg, uniform
    """

    scale = 1.0
    mode = "fan_avg"
    distribution = "uniform"


class He(VarianceScaling):
    """
    Kaiming He (https://arxiv.org/pdf/1502.01852.pdf).
    scale 2, fan_in, normal
    """

    scale = 2.0
    mode = "fan_in"
    distribution = "normal"


HeNormal = He


class HeUniform(He):
    """
    He-init (:class:`He`) but using a uniform distribution.
    scale 2, fan_in, uniform
    """

    distribution = "uniform"


def _compute_fans(shape: Sequence[Dim]):
    """Computes the number of input and output units for a weight shape.

    Args:
      shape: Integer shape tuple or TF tensor shape.

    Returns:
      A tuple of integer scalars (fan_in, fan_out).
    """
    shape = [dim.dimension for dim in shape]
    if len(shape) < 1:  # Just to avoid errors for constants.
        fan_in = fan_out = 1
    elif len(shape) == 1:
        fan_in = fan_out = shape[0]
    elif len(shape) == 2:
        fan_in = shape[0]
        fan_out = shape[1]
    else:
        # Assuming convolution kernels (2D, 3D, or more).
        # kernel shape: (..., input_depth, depth)
        receptive_field_size = 1
        for dim in shape[:-2]:
            receptive_field_size *= dim
        fan_in = shape[-2] * receptive_field_size
        fan_out = shape[-1] * receptive_field_size
    return fan_in, fan_out



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim
from rf_utils import run_model


def test_simple_net_linear():
    batch_dim = Dim(Tensor("batch", [], dtype="int32"))
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim, out_dim = Dim(7, name="in"), Dim(13, name="out")
    extern_data = Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32")

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.linear = rf.Linear(in_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            """
            Forward
            """
            return self.linear(x)

    run_model(_Net, extern_data)



"""
RETURNN frontend (returnn.frontend) utils
"""

from __future__ import annotations
from typing import Callable
import numpy
import torch

import returnn.frontend as rf
from returnn.tensor import Tensor


def run_model(get_model: Callable[[], rf.Module], extern_data: Tensor):
    """run"""
    out_pt = run_model_torch(get_model, extern_data)
    out_tf = run_model_net_dict_tf(get_model, extern_data)
    out_pt, out_tf  # noqa  # TODO ...


def run_model_torch(get_model: Callable[[], rf.Module], extern_data: Tensor) -> Tensor:
    """run"""
    rnd = numpy.random.RandomState(42)
    rf.select_backend_torch()
    _fill_random(extern_data, rnd=rnd, raw_tensor_type=torch.Tensor)
    model = get_model()
    out = model(extern_data)
    assert isinstance(out, Tensor)
    return out


def run_model_net_dict_tf(get_model: Callable[[], rf.Module], extern_data: Tensor):
    """run"""
    get_model, extern_data  # noqa  # TODO
    raise NotImplementedError  # TODO


def _fill_random(x: Tensor, *, rnd: numpy.random.RandomState, raw_tensor_type: type) -> bool:
    """fill. return whether sth was filled"""
    filled = False
    while True:
        have_unfilled = False
        filled_this_round = False

        for dim in x.dims:
            if not dim.dyn_size_ext:
                continue
            if _fill_random(dim.dyn_size_ext, rnd=rnd, raw_tensor_type=raw_tensor_type):
                filled = True
                filled_this_round = True
            if dim.dyn_size_ext.raw_tensor is None:
                have_unfilled = True

        if have_unfilled:
            assert filled_this_round, f"should have filled something, {x}"

        if not have_unfilled:
            break

    if x.raw_tensor is not None:
        if not isinstance(x.raw_tensor, raw_tensor_type):
            x.raw_tensor = None

    if x.raw_tensor is None:
        shape = [dim.get_dim_value() for dim in x.dims]
        if x.dtype.startswith("int"):
            min_val = 0
            max_val = 10
            if x.sparse_dim and x.sparse_dim.dimension is not None:
                max_val = x.sparse_dim.dimension
            values = rnd.randint(min_val, max_val, shape).astype(x.dtype)
        elif x.dtype.startswith("float"):
            values = rnd.uniform(-1, 1, shape).astype(x.dtype)
        else:
            raise NotImplementedError(f"not implemented for {x} dtype {x.dtype}")
        x.raw_tensor = rf.convert_to_tensor(values, dims=x.dims, dtype=x.dtype, sparse_dim=x.sparse_dim).raw_tensor
        filled = True

    assert isinstance(x.raw_tensor, raw_tensor_type)

    return filled



"""
Types
"""

from __future__ import annotations
from typing import Union
import numpy
from returnn.tensor import Tensor, Dim

__all__ = ["RawTensorTypes", "Tensor", "Dim"]

RawTensorTypes = Union[int, float, complex, numpy.number, numpy.ndarray, bool, str]
